from flask import render_template, request, jsonify, current_app, Blueprint
from time import sleep, time
import logging as log
import json as json_module
from sqlalchemy import select, func, and_, delete, distinct, union, alias
from signal import SIGKILL, SIGQUIT, SIGTSTP, SIGCONT
from datetime import datetime

from ..util import package_version, tryupdate, to_dict, flat_list
from .db import db
from .config import IS_SQLITE, UI_OUTPUT_TRUNC, UI_MAX_DISPLAYED_ROW
from ..constants import SIGNAL_CLEAN, SIGNAL_RESTART
from .model import Worker, Signal, Job, Task, Execution, delete_batch, create_worker_create_job, find_flavor
from .api import worker_dao

REFRESH_FLAVOR=60

ui = Blueprint('ui', __name__, url_prefix='/ui')

package_version = package_version()
@ui.route('/')
def index(name=None):
    return render_template('ui.html', name=name, package_version=package_version)

@ui.route('/task/', methods=('GET', 'POST'))
def task():
    if request.method=='POST':
        sortby=request.form.get('sortby','task')
        batch_filter=request.form.get('batch_filter','-')
        worker_filter=request.form.get('worker_filter','-')
        status_filter=request.form.get('status_filter','all')
    else:
        sortby='task'
        batch_filter='-'
        worker_filter='-'
        status_filter='all'
    batch_list = flat_list(db.session.execute(
        select(distinct(alias(union(select(Task.batch),select(Worker.batch)),'batch').table_valued())).order_by('batch')
    ))
    worker_list = flat_list(db.session.query(
        distinct(Worker.name)
    ))
    
    #print('batch_list',repr(batch_list))
    return render_template('task.html',package_version=package_version,
                           sortby=sortby,
                           batch_filter=batch_filter,
                           worker_filter=worker_filter,
                           status_filter=status_filter,
                           batch_list=batch_list,
                           worker_list=worker_list)

@ui.route('/batch/')
def batch():
    return render_template('batch.html', package_version=package_version)



@ui.route('/get/')
def handle_get():
    json = request.args
    if 'delay' in json:
        sleep(int(json['delay']))
    if json['object']=='workers':
        log.info('sending workers')
        return jsonify({
            'workers':list([tryupdate(dict(row),'stats',json_module.loads,row.stats) for row in db.session.execute(
                '''SELECT 
                    worker_id,
                    name, 
                    batch,
                    status, 
                    concurrency,
                    prefetch,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND
                                                                        execution.status='accepted') as accepted,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND 
                                                                        execution.status='running') as running,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND 
                                                                        execution.status='succeeded') as succeeded,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id AND 
                                                                        execution.status='failed') as failed,
                    (SELECT count(execution_id) FROM execution WHERE execution.worker_id=worker.worker_id) as total,
                    load,
                    memory,
                    stats,
                    flavor
                FROM worker
                ORDER BY worker.batch,worker.name''')]),
            'tasks_per_status': dict([row for row in db.session.execute(
                '''SELECT status,count(task_id) as count FROM task GROUP BY status'''
            )]) })

    elif json['object']=='tasks':
        sortby = json.get('sortby', None)
        status = json.get('status', None)
        worker = json.get('worker', None)
        batch = json.get('batch',None)

        log.warning(f"sending task ordered by {sortby} filtered by status {status} for worker {worker}")


        if sortby=='worker':
            sort_clause='ORDER BY worker.name, task.task_id DESC'
        elif sortby=='batch':
            sort_clause='ORDER BY task.batch, task.task_id DESC'
        else:
            sort_clause='ORDER BY task.task_id DESC'
        
        where_clauses = []

        if status is not None:
            where_clauses.append(f"task.status='{status}'")

        if worker is not None:
            where_clauses.append(f"worker.name='{worker}'")

        if batch is not None:
            where_clauses.append(f"task.batch='{batch}'")

        where_clause = f'''WHERE {' AND '.join(where_clauses)}''' if where_clauses else ''

        if IS_SQLITE:
            trunc_output=f'SUBSTR(execution.output,-{UI_OUTPUT_TRUNC},{UI_OUTPUT_TRUNC})'
            trunc_error=f'SUBSTR(execution.error,-{UI_OUTPUT_TRUNC},{UI_OUTPUT_TRUNC})'
        else:
            trunc_output=f'RIGHT(execution.output,{UI_OUTPUT_TRUNC})'
            trunc_error=f'RIGHT(execution.error,{UI_OUTPUT_TRUNC})'
        

        #task_list = list([list(map(lambda x : str(x) if type(x)==type(datetime.utcnow()) else x ,row)) for row in db.session.execute(
        task_list = list([dict(row) for row in db.session.execute(
        f'''SELECT
        task.task_id,
        task.name,
        worker.name as worker_name,
        task.batch,
        execution.creation_date,
        execution.modification_date,
        execution.execution_id,
        {trunc_output} as output,
        {trunc_error} as error,
        task.command,
        execution.worker_id,
        task.status
        FROM task 
        LEFT JOIN execution ON (task.task_id=execution.task_id AND latest)
        LEFT JOIN worker ON execution.worker_id=worker.worker_id 
        {where_clause}
        {sort_clause}
        LIMIT {UI_MAX_DISPLAYED_ROW}
        '''
        )])
        
        detailed_tasks = json.getlist('detailed_tasks[]')
        if detailed_tasks:
            log.warning(f"detailing tasks {detailed_tasks}")
            for detailed_task in db.session.execute(f"""
                SELECT execution_id,output,error FROM execution 
                WHERE execution_id IN ({','.join([str(eid) for eid in detailed_tasks])})"""):
                for task in task_list:
                    if task['execution_id']==detailed_task['execution_id']:
                        task['output']=detailed_task['output']
                        task['error']=detailed_task['error']
                        break

        batch_list = flat_list(db.session.execute(
            select(distinct(alias(union(select(Task.batch),select(Worker.batch)),'batch').table_valued())).order_by('batch')
        ))
        worker_list = flat_list(db.session.query(
            distinct(Worker.name)
        ))

        return jsonify({'tasks':task_list, 'batch_list':batch_list, 'worker_list':worker_list})
        
    elif json['object'] == 'batch':
        log.info('sending batch')
        if IS_SQLITE:
            duration_query='(JULIANDAY(e1.modification_date)-JULIANDAY(e1.creation_date))*24'
            worker_query='''SELECT batch,GROUP_CONCAT(name,',') as workers FROM worker GROUP BY batch ORDER BY name'''
        else:
            duration_query='EXTRACT ( EPOCH FROM (e1.modification_date-e1.creation_date)/3600 )'
            worker_query='''SELECT batch,STRING_AGG(name,',') as workers FROM (SELECT * FROM worker ORDER BY name) w GROUP BY batch'''
        batch_query=f'''SELECT * FROM (
    SELECT batch,status,COUNT(task_id) as count,MAX(duration) as max,MIN(duration) as min, AVG(duration) as avg FROM (
        SELECT {duration_query} as duration, e1.task_id, e1.status,task.batch FROM execution e1 JOIN task ON (
            task.task_id=e1.task_id AND e1.latest
        )
    ) AS e2 GROUP BY batch,status
    UNION	 
    SELECT batch,status, COUNT(task_id),NULL,NULL,NULL 
    FROM task WHERE task_id NOT IN (SELECT task_id FROM execution) GROUP BY batch,status
) AS b ORDER BY batch, status'''
        return jsonify({'batches':list([dict(row) for row in db.session.execute(batch_query)]),
                        'workers': list([dict(row) for row in db.session.execute(worker_query)])})
    

@ui.route('/flavors/')
def flavors():
    """Provide flavor list to the UI to make worker deployment through the UI more user friendly.
    Take a object with a .date attribute containing the last time of refresh with an epoch (int in seconds)"""
    date = int(request.args['date'])
    epoch = int(time())
    if epoch-date>REFRESH_FLAVOR:
        return jsonify({'list':find_flavor(session=db.session,limit=None),'date':epoch})
    else:
        return '"up to date"'


#@socketio.on('change_batch')
@ui.route('/change_batch')
def handle_change_batch():
    json = request.args
    Worker.query.filter(Worker.worker_id==json['worker_id']).update(
        {Worker.batch:json['batch_name'] or None})
    db.session.commit()
    return '"ok"'


#@socketio.on('concurrency_change')
@ui.route('/concurrency_change')
def handle_concurrency_change():
    json = request.args
    worker_id = json['id']
    change = json['change']
    log.info(f'changing concurrency for worker {worker_id}: {change}')
    if IS_SQLITE:
        log.info('Using sqlite SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.concurrency: func.max(Worker.concurrency+change,0)})
    else:
        log.info('Using standard SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.concurrency: func.greatest(Worker.concurrency+change,0)})
    db.session.commit()
    return '"ok"'

#@socketio.on('prefetch_change')
@ui.route('/prefetch_change')
def handle_prefetch_change():
    json = request.args
    worker_id = json['id']
    change = json['change']
    log.info(f'changing prefetch for worker {worker_id}: {change}')
    if IS_SQLITE:
        log.info('Using sqlite SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.prefetch: func.max(Worker.prefetch+change,0)})
    else:
        log.info('Using standard SQL')
        Worker.query.filter(Worker.worker_id==worker_id).update(
            {Worker.prefetch: func.greatest(Worker.prefetch+change,0)})
    db.session.commit()
    return '"ok"'

@ui.route('/pause_unpause_worker')
def handle_pause_unpause_worker():
    json = request.args
    worker_id = json['id']
    status = json['status']
    Worker.query.filter(Worker.worker_id==worker_id).update({Worker.status:status})
    log.info(f'changing status for worker {worker_id}: {status}')
    db.session.commit()
    return '"ok"'


@ui.route('/clean_worker')
def handle_clean_worker():
    json = request.args
    worker_id = json['worker_id']
    db.session.add(Signal(execution_id=None, worker_id=worker_id, signal=SIGNAL_CLEAN))
    db.session.commit()
    log.info(f'sending clean signal for worker {worker_id}')
    return '"ok"'

@ui.route('/restart_worker')
def handle_restart_worker():
    json = request.args
    worker_id = json['worker_id']
    db.session.add(Signal(execution_id=None, worker_id=worker_id, signal=SIGNAL_RESTART))
    db.session.commit()
    log.info(f'sending restart signal for worker {worker_id}')
    return '"ok"'

#@socketio.on('create_worker')
@ui.route('/create_worker')
def handle_create_worker():
    json = request.args
    concurrency = int(json['concurrency'])
    flavor = json['flavor']
    if not flavor:
        return jsonify(error='Flavor must be specified')
        #return None
    region = json['region'].strip()
    provider = json['provider'].strip()
    if not region:
        return jsonify(error='Region must be specified')
        #return None
    batch = json['batch'] or None
    prefetch = int(json['prefetch'])
    number = int(json['number'])
    create_worker_create_job(concurrency=concurrency, prefetch=prefetch, batch=batch,
                             flavor=flavor, region=region, provider=provider, session=db.session,
                             number=number)
    #for _ in range(number):
    #    db.session.add(
    #        Job(target='', 
    #            action='worker_create', 
    #            args={
    #                'concurrency': concurrency, 
    #                'prefetch':prefetch,
    #                'flavor':flavor,
    #                'region':region,
    #                'provider': provider,
    #                'batch':batch
    #            }
    #        )
    #    )
    #db.session.commit()
    return '"ok"'

#@socketio.on('batch_action')
@ui.route('/batch/action')
def handle_batch_action():
    """Gathering all the action dealing with batch like pause, break, stop, clear, go."""
    json = request.args
    if json['action'] in ['stop','break','pause','simple pause','pause only batch']:
        #Same function as in the API set all workers affected to this batch to running and can also interrupt the running tasks with signal 3 and 9
        name=json['name']
        for w in Worker.query.filter(Worker.batch==name):
                w.status = 'paused'
        if json['action']=='break':
            signal = SIGKILL
        elif json['action']=='stop':
            signal = SIGQUIT
        elif json['action']=='simple pause':
            signal = 0
        elif json['action']=='pause':
            #TODO: make a more efficient query
            for t in Task.query.filter(and_(
                        Task.batch==name,
                        Task.status.in_(['running','accepted']))):
                t.status = 'paused'
            signal = SIGTSTP
            db.session.commit()
        log.warning(f'Sending signal {signal} to executions for batch {name}')
        for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                        Execution.status=='running',
                                        Task.batch==name)):
            log.warning(f'Sending signal {signal} to execution {e.execution_id}')
            db.session.add(Signal(e.execution_id, e.worker_id, signal ))
        db.session.commit()
        log.warning('result pause :Ok')
    elif json['action'] in ['go','simple go']: 
        #Same function as in the API (re)set all workers affected to this batch to running
        name=json['name']
        """(re)set all workers affected to this batch to running"""
        for w in Worker.query.filter(Worker.batch==name):
            w.status = 'running'
        if json['action']=='go':
            signal=SIGCONT
            for t in Task.query.filter(Task.batch==name):
                if t.status == 'paused':
                    t.status='running'
            for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                            Execution.status=='running',
                                            Task.batch==name)):
                log.warning(f'Sending signal {signal} to execution {e.execution_id}')
                db.session.add(Signal(e.execution_id, e.worker_id, signal ))
        db.session.commit()
        log.warning('result go : Ok')
    elif json['action']=='clear':
        #Same function as in the API clear() Delete all tasks and executions and recruiters for this batch
        name=json['name']
        """Delete all tasks and executions for this batch"""
        delete_batch(name, session=db.session)
        log.warning(f'result clear batch {name}: Ok ')
    return '"ok"'

#@socketio.on('task_action')
@ui.route('/task/action')
def handle_task_action():
    """Gathering all the action dealing with task like break, stop, delete, modify, restart"""
    #The code essentially is from the API code with a few modifications
    json = request.args
    task=json['task_id']
    if json['action'] in ['break','stop','pause','resume']: 
        #A signal 3 or 9 is created and causes only the interruption of the task with id same structure as in the API 
        task=json['task_id']
        for t in Task.query.filter(Task.task_id==task):
            if json['action'] == 'break':
                type='break'
                signal = SIGKILL
            elif json['action'] == 'stop':
                signal = SIGQUIT
                type='stop'
            elif json['action'] == 'pause':
                signal = SIGTSTP
                type='pause'
                t.status='paused'
            elif json['action'] == 'resume':
                signal = SIGCONT
                type='resume'
                t.status='running'
            log.warning(f'Sending signal {signal} to executions for task {task}')
            for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                            Execution.status=='running',
                                            Task.task_id==task)):
                log.warning(f'Sending signal {signal} to execution {e.execution_id}')
                db.session.add(Signal(e.execution_id, e.worker_id, signal ))
        db.session.commit()
        log.warning(f'result {type} : Ok')
    elif json['action']=='delete': 
        #Delete the task in the data base
        for t in Task.query.filter(Task.task_id==task):
            db.session.delete(t)
        db.session.commit()
        log.warning('result delete: Ok')
    elif json['action']=='modify': 
        #Changing the command for a task in the data base and moving it in the task queue. It doesn't create a new task.
        now = datetime.utcnow()
        for t in Task.query.filter(Task.task_id==task):
            t.command =json["modification"]
            if t.status not in ['pending','assigned','accepted']:
                for e in t.executions:
                    if e.status=='running':
                        e.status='failed'
                        db.session.add(Signal(e.execution_id, e.worker_id, SIGKILL))
                t.status='pending' 
                t.status_date=now
                t.modification_date=now          
        db.session.commit()
        log.warning('result modify : Ok')
    elif json['action']=='restart': 
        #Relaunching the execution of a task.
        now = datetime.utcnow()
        for t in Task.query.filter(Task.task_id==task):
            t.status='pending'
            t.status_date=now
            t.modification_date=now  
        db.session.commit()
        log.warning('result restart : Ok')
    return '"ok"'

#@socketio.on('delete_worker') #Delete a worker.
@ui.route('/delete_worker')
def delete_worker():
    """Delete a worker in db"""
    json = request.args
    worker_dao.delete(json['worker_id'], session=db.session)
    return '"ok"'
    

#@socketio.on('jobs')
@ui.route('/jobs')
def handle_jobs():
    """Provide UI with job list"""
    return jsonify(jobs = [to_dict(job) for job in db.session.query(Job).order_by(Job.job_id.desc()).all()])

@ui.route('/delete_job')
def delete_job():
    """Delete a job in db"""
    json = request.args
    job = db.session.query(Job).get(json['job_id'])
    if job:
        db.session.delete(job)
        db.session.commit()
    else:
        log.warning(f"Job {json['job_id']} already deleted")
    return '"ok"'

@ui.route('/delete_jobs')
def delete_jobs():
    """Delete terminated jobs"""
    if db.session.execute(
            select(func.count()).select_from(Job).where(Job.status=='succeeded')
            ).scalar_one()>0:
        log.warning('Here')
        db.session.execute(delete(Job).where(Job.status=='succeeded'))
    elif db.session.execute(
            select(func.count()).select_from(Job).where(Job.status=='failed')
            ).scalar_one()>0:
        log.warning('There')
        db.session.execute(delete(Job).where(Job.status=='failed'))
    elif db.session.execute(
            select(func.count()).select_from(Job).where(Job.status=='pending')
            ).scalar_one()>0:
        db.session.execute(delete(Job).where(Job.status=='pending'))
    else:
        return '"nothing to do"'
    db.session.commit()
    return '"ok"'

@ui.route('/restart_job')
def restart_job():
    """Restart a job in db"""
    json = request.args
    job = db.session.query(Job).get(json['job_id'])
    if job:
        job.status = 'pending'
        job.retry = 0
        db.session.commit()
    else:
        log.warning(f"Job {json['job_id']} already deleted")
    return '"ok"'

#ns = api.namespace('live', description='Very basic API for the UI')

@ui.route('/ping')
def live_ping():
    """Simply return ok"""
    return '"ok"'


"""Modifying defaults and adding detail in workers

Revision ID: 00ca4479c7e1
Revises: 
Create Date: 2023-12-08 15:39:55.492553

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '00ca4479c7e1'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('task', schema=None) as batch_op:
        batch_op.add_column(sa.Column('retry', sa.Integer(), nullable=False,server_default="0"))
        batch_op.alter_column('batch',
               existing_type=sa.VARCHAR(),
               nullable=False)

    with op.batch_alter_table('worker', schema=None) as batch_op:
        batch_op.add_column(sa.Column('task_properties', sa.String(), nullable=False,server_default='{}'))
        batch_op.add_column(sa.Column('flavor', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('region', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('provider', sa.String(), nullable=True))
        batch_op.alter_column('concurrency',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('batch',
               existing_type=sa.VARCHAR(),
               nullable=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('worker', schema=None) as batch_op:
        batch_op.alter_column('batch',
               existing_type=sa.VARCHAR(),
               nullable=True)
        batch_op.alter_column('concurrency',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.drop_column('provider')
        batch_op.drop_column('region')
        batch_op.drop_column('flavor')
        batch_op.drop_column('task_properties')

    with op.batch_alter_table('task', schema=None) as batch_op:
        batch_op.alter_column('batch',
               existing_type=sa.VARCHAR(),
               nullable=True)
        batch_op.drop_column('retry')

    # ### end Alembic commands ###



import os
from ..constants import DEFAULT_SERVER_CONF

if 'FLASK_APP' not in os.environ and 'SCITQ_PRODUCTION' not in os.environ:
    import dotenv
    dotenv.load_dotenv(DEFAULT_SERVER_CONF, override=False)

from socket import gethostname
from ..util import check_dir, package_path
from logging.config import dictConfig
from ..default_settings import SQLALCHEMY_POOL_SIZE, SQLALCHEMY_DATABASE_URI




MAIN_THREAD_SLEEP = 5
WORKER_OFFLINE_DELAY = 15
SCITQ_SERVER = os.environ.get('SCITQ_SERVER',None)

WORKER_CREATE = f'cd {package_path("ansible","playbooks")} && ansible-playbook deploy_one_vm.yaml --extra-vars \
"nodename={{hostname}} concurrency={{concurrency}} status=running flavor={{flavor}} \
region={{region}} provider={{provider}} tags={{tags}}"'

if SCITQ_SERVER is not None:
    WORKER_CREATE = WORKER_CREATE[:-1] + f' target={SCITQ_SERVER}"'
    SCITQ_SHORTNAME = SCITQ_SERVER.split('.')[0]
else:
    SCITQ_SHORTNAME = gethostname().split('.')[0]
WORKER_DELETE = os.environ.get('WORKER_DELETE',
    f'cd {package_path("ansible","playbooks")} && ansible-playbook destroy_vm.yaml --extra-vars "nodename={{hostname}}"')
SERVER_CRASH_WORKER_RECOVERY = os.environ.get('SERVER_CRASH_WORKER_RECOVERY',
    f'cd {package_path("ansible","playbooks")} && ansible-playbook check_after_reboot.yaml')
WORKER_IDLE_CALLBACK = os.environ.get('WORKER_IDLE_CALLBACK',WORKER_DELETE)
WORKER_CREATE_CONCURRENCY = 10
WORKER_CREATE_RETRY=2
WORKER_CREATE_RETRY_SLEEP=30
UI_OUTPUT_TRUNC=100
UI_MAX_DISPLAYED_ROW = 500
WORKER_DESTROY_RETRY=2
DEFAULT_BATCH = 'Default'
TERMINATE_TIMEOUT = 20
KILL_TIMEOUT = 30
JOB_MAX_LIFETIME = 600

def _(x):
    """a fail-free shortcut to os.environ.get to import an env variable"""
    return os.environ.get(x,'')

def _num(x, default=None):
    """a failing shortcut to os.environ.get to import an env variable supposed to be numerical"""
    if x in os.environ:
        return float(os.environ[x])
    return default

AZURE_REGIONS=_('AZURE_REGIONS')
AZURE_CPUQUOTAS=_('AZURE_CPUQUOTAS')
AZURE_SUBSCRIPTION_ID=_('AZURE_SUBSCRIPTION_ID')
AZURE_CLIENT_ID=_('AZURE_CLIENT_ID')
AZURE_SECRET=_('AZURE_SECRET')
AZURE_TENANT=_('AZURE_TENANT')
PREFERRED_REGIONS=dict([preference.split(':') for preference in _('PREFERRED_REGIONS').split(',') if ':' in preference])

OVH_REGIONS=_('OVH_REGIONS')
OVH_CPUQUOTAS=_('OVH_CPUQUOTAS')
OVH_APPLICATIONKEY    = _('OVH_APPLICATIONKEY') 
OVH_APPLICATIONSECRET = _('OVH_APPLICATIONSECRET')
OVH_CONSUMERKEY       = _('OVH_CONSUMERKEY')
OS_PROJECT_ID         = _('OS_PROJECT_ID')

EVICTION_ACTION=_('EVICTION_ACTION')
EVICTION_COST_MARGIN=_num('EVICTION_COST_MARGIN', default=10)

def get_quotas(provider=None):
    if provider=='ovh':
        return dict(zip(OVH_REGIONS.split(),map(int,OVH_CPUQUOTAS.split())))
    elif provider=='azure':
        return dict(zip(AZURE_REGIONS.split(),map(int,AZURE_CPUQUOTAS.split())))
    elif provider is None:
        quotas = {}
        for provider in ['ovh','azure']:
            for region, cpuquota in get_quotas(provider).items():
                quotas[(provider,region)]=cpuquota
        return quotas
    else:
        raise RuntimeError(f'Cannot give quotas for unknown provider: {provider}')

def setup_log():
    """Setting up log must occur only in specific contexts"""
    if os.environ.get('QUEUE_PROCESS') and os.environ.get('QUEUE_LOG_FILE'):
        check_dir(os.environ.get('QUEUE_LOG_FILE'))
        dictConfig({
            'version': 1,
            'formatters': {'default': {
                'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
            }},
            'handlers': {'wsgi': {
                'class': 'logging.StreamHandler',
                'stream': 'ext://flask.logging.wsgi_errors_stream',
                'formatter': 'default'
            }, "file": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "default",
                "filename": os.environ.get('QUEUE_LOG_FILE'),
                "maxBytes": int(os.environ.get('QUEUE_LOG_FILE_MAX_SIZE',
                    os.environ.get('LOG_FILE_MAX_SIZE',"10000000"))),
                "backupCount": int(os.environ.get('QUEUE_LOG_FILE_KEEP',
                    os.environ.get('LOG_FILE_KEEP',"3")))
            }},
            'root': {
                'level': os.environ.get('LOG_LEVEL',"INFO"),
                'handlers': ['wsgi' if 'DEBUG' in os.environ else 'file']
            }
        })
    else:
        check_dir(os.environ.get('LOG_FILE',"/tmp/scitq.log"))
        dictConfig({
            'version': 1,
            'formatters': {'default': {
                'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
            }},
            'handlers': {'wsgi': {
                'class': 'logging.StreamHandler',
                'stream': 'ext://flask.logging.wsgi_errors_stream',
                'formatter': 'default'
            }, "file": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "default",
                "filename": os.environ.get('LOG_FILE',"/tmp/scitq.log"),
                "maxBytes": int(os.environ.get('LOG_FILE_MAX_SIZE',"10000000")),
                "backupCount": int(os.environ.get('LOG_FILE_KEEP',"3"))
            }},
            'root': {
                'level': os.environ.get('LOG_LEVEL',"INFO"),
                'handlers': ['wsgi' if 'DEBUG' in os.environ else 'file']
            }
        })

IS_SQLITE = 'sqlite' in SQLALCHEMY_DATABASE_URI




"""Add timeout to tasks

Revision ID: c225e89357b9
Revises: 00ca4479c7e1
Create Date: 2024-01-25 10:25:55.535129

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c225e89357b9'
down_revision = '00ca4479c7e1'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('task', schema=None) as batch_op:
        batch_op.add_column(sa.Column('status_date', sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column('download_timeout', sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column('run_timeout', sa.Integer(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('task', schema=None) as batch_op:
        batch_op.drop_column('run_timeout')
        batch_op.drop_column('download_timeout')
        batch_op.drop_column('status_date')

    # ### end Alembic commands ###



#!/usr/bin/env python
from .lib  import Server
from subprocess import run
import argparse
import sys
from datetime import datetime
import os 
from .util import package_path, package_version
import shutil
import random
from .debug import Debugger
from .constants import DEFAULT_SERVER_CONF, DEFAULT_WORKER_CONF, TASK_STATUS, EXECUTION_STATUS,\
    FLAVOR_DEFAULT_LIMIT, FLAVOR_DEFAULT_EVICTION
from signal import SIGKILL, SIGCONT, SIGQUIT, SIGTSTP
import dotenv
from tabulate import tabulate

MAX_LENGTH_STR=50
DEFAULT_SERVER = os.getenv('SCITQ_SERVER','127.0.0.1')
DEFAULT_ANSIBLE_INVENTORY = '/etc/ansible/inventory'
DEFAULT_PROVIDER='ovh'
OLD_SQLITE_INVENTORY_MD5='723562df744fc082fa7fc03a41d10c3c'

def converter(x,long): 
    """A small conversion function for items in lists"""
    if type(x)==datetime:
        x=str(x)
    elif type(x)==str and not long:
        if len(x)> MAX_LENGTH_STR :
            x=x[:MAX_LENGTH_STR-1] +'...'
    elif type(x)==list:
        x=','.join([str(element) for element in x])
    elif type(x)==bool:
        x='X' if x else ''
    elif x is None:
        x=''
    return x


def __list_print(item_list, retained_columns, headers, long=False):
    """A small internal function to format a list of items (i.e. tasks or workers)"""
    print(tabulate(
        [[converter(item.get(key,None),long) for key in retained_columns]
            for item in item_list],
        headers=headers,tablefmt ="plain"
        ))

def main():
    parser = argparse.ArgumentParser()
    
    parser.add_argument('-s','--server', help=F"Define a server, by default {DEFAULT_SERVER} this default value comes from the SCITQ_SERVER environment variable",type=str,default=DEFAULT_SERVER)
    parser.add_argument('-t','--timeout', type=int,
            help='Specify the get timeout for complex queries (default to 150s)' )
    parser.add_argument('--version', action='store_true',
            help="Print the version and exit")

    subparser=parser.add_subparsers(help='sub-command help',dest='object')
    worker_parser = subparser.add_parser('worker', help='The following options will only concern workers')
    subsubparser=worker_parser.add_subparsers(dest='action')
    list_worker_parser=subsubparser.add_parser('list', help='List all workers')
    list_worker_parser.add_argument('-b','--batch',help="Give you a list of workers according to his name",type=str,default='')
    list_worker_parser.add_argument('-S','--status',help="give you a list of workers according to his status",type=str,default='',choices=['','running','paused','offline','failed'])
    list_worker_parser.add_argument('-H','--no-header',help='Do not print headers',action='store_true')
    list_worker_parser.add_argument('-L','--long',help='Print the entire command',action='store_true')

    deploy_parser=subsubparser.add_parser('deploy',help="Deploy a new worker")
    deploy_parser.add_argument('-C','--concurrency',help="Define how many tasks the worker can do at once (default to 1)",type=int,default=1)
    deploy_parser.add_argument('-p','--prefetch',help="Define how many tasks should be prefetched (default to 0)",type=int,default=0)
    deploy_parser.add_argument('-f','--flavor',required=True,help="Define which flavor/model should be ordered (MANDATORY)",type=str)
    deploy_parser.add_argument('-r','--region',required=True,help="Define in which provider region to order (MANDATORY)",type=str)
    deploy_parser.add_argument('-P','--provider',required=True,help=f"Define from which provider rent the worker (default to {DEFAULT_PROVIDER})",
                               type=str, default=DEFAULT_PROVIDER)
    deploy_parser.add_argument('-N','--number',help="Define the number of workers you want to create (default to 1)",type=int, default=1)
    deploy_parser.add_argument('-b','--batch',help="Define the batch that should fit the tasks'one (default to None, default batch)",type=str,default=None)

    worker_delete_parser=subsubparser.add_parser('delete',help="Delete a worker, possibly terminating the VM")
    worker_delete_parser.add_argument('-n','--name',help="The name of the worker to delete (MANDATORY if -i not used)",type=str, default=None)
    worker_delete_parser.add_argument('-i','--id',help="The id of the worker to delete (MANDATORY if -n is not used)",type=int, default=None)

    worker_update_subparser = subsubparser.add_parser('update', help='Modify some worker')
    worker_update_subparser.add_argument('-i','--id', help='The ID of the worker to modify (MANDATORY if -n is not used)', type=int, default=None)
    worker_update_subparser.add_argument('-n','--name', help='The name of the worker to modify (MANDATORY if -i is not used)', type=str, default=None)
    worker_update_subparser.add_argument('-S','--status', help='The new status of the worker', type=str, default=None)
    worker_update_subparser.add_argument('-b','--batch', help='The new batch of the worker', type=str, default=None)
    worker_update_subparser.add_argument('-C','--concurrency', help='The new concurrency of the worker', type=str, default=None)
    worker_update_subparser.add_argument('-p','--prefetch',help="Define how many tasks should be prefetched",type=int,default=None)
    worker_update_subparser.add_argument('-f','--flavor',help="Define which flavor/model should be ordered",type=str,default=None)
    

    batch_parser = subparser.add_parser('batch', help='The following options will only concern batches')
    subsubparser=batch_parser.add_subparsers(dest='action')
    list_parser=subsubparser.add_parser('list',help='List batches')

    pause_parser=subsubparser.add_parser('stop',help='Stop (pause) a batch')
    pause_parser.add_argument('-n','--name', help='Target the batch with his name. Accept a name or also a list of name by using -n name1 -n name2',action='append',required=True)
    option_group=pause_parser.add_mutually_exclusive_group()
    option_group.add_argument('-N','--number',help='Create a signal if number equals 3, 9, 20',default=0,type=int,choices=[0,3,9,20])
    option_group.add_argument('--term',help='Send the TERM signal (same as -N 3)',action='store_true')
    option_group.add_argument('--kill',help='Send the KILL signal (same as -N 9)',action='store_true')
    option_group.add_argument('--pause',help='Send the SIGTSTP (pause) signal (same as -N 20)',action='store_true')


    go_parser=subsubparser.add_parser('go',help='Continue (relaunch) the execution of a batch ')
    go_parser.add_argument('-n','--name',help='Target the batch with his name. Accept a name or also a list of name by using -n name1 -n name2',action='append',required=True)
    option_group=go_parser.add_mutually_exclusive_group()
    option_group.add_argument('-N','--number',help='Create a signal if number equals 18',default=0,type=int,choices=[0,18])
    option_group.add_argument('--cont',help='Send the SIGCONT (continue) signal (same as -N 18)',action='store_true')

    clear_parser=subsubparser.add_parser('delete',help='Delete (clear) a batch')
    clear_parser.add_argument('-n','--name', help='Target the batch with his name. Accept a name or also a list of name by using -n name1 -n name2',action='append',required=True)

    task_parser = subparser.add_parser('task', help='The following options will only concern tasks')
    subsubparser=task_parser.add_subparsers(dest='action')
    list_task_parser= subsubparser.add_parser('list',help='List all tasks')
    list_task_parser.add_argument('-S','--status',help='Give you a list of tasks according to his status',type=str,choices=TASK_STATUS,default='')
    list_task_parser.add_argument('-b','--batch',help='Give you a list of tasks according to his batch',type=str,default='')
    list_task_parser.add_argument('-H','--no-header',help='Do not print the headers',action='store_true')
    list_task_parser.add_argument('-L','--long',help='Print the entire command',action='store_true')

    go_task_parser= subsubparser.add_parser('relaunch',help='Relaunch a task')
    option_group=go_task_parser.add_mutually_exclusive_group()
    option_group.add_argument('-i','--id',help='Target a task by his id (MANDATORY if -n is not used)',type=int)
    option_group.add_argument('-n','--name', help='The name of the task to delete (MANDATORY if -i is not used)', type=str)
    
    output_task_parser= subsubparser.add_parser('output',help='Show the output/error for a task')
    option_group=output_task_parser.add_mutually_exclusive_group()
    option_group.add_argument('-i','--id',help='Target a task by his id (MANDATORY if -n is not used)',type=int)
    option_group.add_argument('-n','--name', help='The name of the task to delete (MANDATORY if -i is not used)', type=str)
    option_group=output_task_parser.add_mutually_exclusive_group()
    option_group.add_argument('-o','--output',help='Show only the output for a task',action='store_true')
    option_group.add_argument('-e','--error',help='Show only the error for a task',action='store_true')

    task_delete_subparser = subsubparser.add_parser('delete', help='Delete some task')
    option_group=task_delete_subparser.add_mutually_exclusive_group()
    option_group.add_argument('-i','--id',help='Target a task by his id (MANDATORY if -n is not used)',type=int)
    option_group.add_argument('-n','--name', help='The name of the task to delete (MANDATORY if -i is not used)', type=str)
    
    task_update_subparser = subsubparser.add_parser('update', help='Modify some task')
    option_group=task_update_subparser.add_mutually_exclusive_group()
    option_group.add_argument('-i','--id',help='Target a task by his id (MANDATORY if -n is not used)',type=int)
    option_group.add_argument('-n','--name', help='The name of the task to delete (MANDATORY if -i is not used)', type=str)
    task_update_subparser.add_argument('-N','--new-name', help='The new name of the task', type=str, default=None)
    task_update_subparser.add_argument('-c','--command', help='The new command of the task', type=str, default=None)
    task_update_subparser.add_argument('-S','--status', help='The new status of the task', type=str, default=None)
    task_update_subparser.add_argument('-b','--batch', help='The new batch of the task', type=str, default=None)
    task_update_subparser.add_argument('-d','--docker', help='The new docker for the task', type=str, default=None)
    task_update_subparser.add_argument('-O','--option', help='The new docker option for the task', type=str, default=None)
    task_update_subparser.add_argument('-j','--input', help='The new input for the task', type=str, default=None)
    task_update_subparser.add_argument('-o','--output', help='The new output for the task', type=str, default=None)
    task_update_subparser.add_argument('-R','--requirements', help='A new space separated required task ids', type=str, default=None)
    task_update_subparser.add_argument('--run-timeout', help='Change the run timeout (in seconds) for this task', type=int, default=None)
    task_update_subparser.add_argument('--download-timeout', help='Change the download timeout (in seconds) for this task', type=int, default=None)
    
    ansible_parser = subparser.add_parser('ansible', help='The following options are to work with ansible subcode')
    subsubparser=ansible_parser.add_subparsers(dest='action')
    ansible_path_parser=subsubparser.add_parser('path',help='Return the path of scitq Ansible playbooks')
    ansible_install_parser=subsubparser.add_parser('install',help='Install scitq Ansible inventory files')
    ansible_install_parser.add_argument('-p','--path', help=f'specify install path (default to {DEFAULT_ANSIBLE_INVENTORY})',
         type=str, default=DEFAULT_ANSIBLE_INVENTORY)
    ansible_inventory_parser=subsubparser.add_parser('inventory',help='Deprecated, use worker list instead')

    debug_parser = subparser.add_parser('debug', help='The following options help debuging a new task')
    subsubparser=debug_parser.add_subparsers(dest='action')
    debug_run_parser=subsubparser.add_parser('run',help='Run a debugging task (pick randomly if -i or -n is not specified)')
    debug_run_parser.add_argument('-b','--batch',help='Pick task in this batch',type=str, default=None)
    debug_run_parser.add_argument('-i','--id',help='Use the task with this id',type=int, default=None)
    debug_run_parser.add_argument('-n','--name', help='Use the task with this id', type=str, default=None)
    debug_run_parser.add_argument('-r','--retry', help='Do not re-download input and resources, it is just a retry', action='store_true')
    debug_run_parser.add_argument('--no-resource', help='Do not re-download resources, but re-download inputs', action='store_true')
    debug_run_parser.add_argument('-c','--conf', help=f'Use this environment file to set environment (default to {DEFAULT_SERVER_CONF})', type=str, default=DEFAULT_SERVER_CONF)
    debug_run_parser.add_argument('-w','--worker-conf', help=f'Use this environment file to set extra environment (default to {DEFAULT_WORKER_CONF})', 
                                  type=str, default=DEFAULT_WORKER_CONF)
    
    recruiter_parser = subparser.add_parser('recruiter', help='Do some action about recruiter objects (recruiters automatically deploy workers in a batch when needed)')
    subsubparser=recruiter_parser.add_subparsers(dest='action')

    list_recruiter_parser=subsubparser.add_parser('list', help='List all recruiter')
    list_recruiter_parser.add_argument('-b','--batch',help="Filter recruiter for this batch",type=str,default=None)
    list_recruiter_parser.add_argument('-L','--long',help='Print all recruiter details',action='store_true')
    list_recruiter_parser.add_argument('-H','--no-header',help='Do not print headers',action='store_true')

    recruiter_create_parser = subsubparser.add_parser('create', help='Create a new recruiter (or override a recruiter of the same rank for this batch)')
    recruiter_create_parser.add_argument('-b','--batch',type=str,required=True,help='batch in which workers are deployed')
    recruiter_create_parser.add_argument('-n','--rank',type=int,help='rank in which recruiter is tried (unique per batch), default to 1',default=1)
    recruiter_create_parser.add_argument('-f','--flavor',type=str,required=True,help='flavor (type of instance) of worker needed')
    recruiter_create_parser.add_argument('-r','--region',type=str,help='worker region of provenance',default=None)
    recruiter_create_parser.add_argument('-P','--provider',type=str,help='worker provider',default=None)
    recruiter_create_parser.add_argument('-C','--concurrency',type=int,required=True,help='what should be the concurrency setting of recruited workers')
    recruiter_create_parser.add_argument('-p','--prefetch',type=int,help='what should be the prefetch setting of recruited workers (default to 0)',default=None)
    recruiter_create_parser.add_argument('-T','--tasks-per-worker',type=int,help='how many workers should be recruited (default to concurrency, but if you want each worker to work two rounds, set it to twice the concurrency)',default=None)
    recruiter_create_parser.add_argument('-m','--minimum-tasks',type=int,help='prevent the recruiter from triggering before this minimal number of tasks is reached',default=None)
    recruiter_create_parser.add_argument('-W','--maximum-workers',type=int,help='prevent the recruiter from recruiting more than this number of worker',default=None)
    
    recruiter_modify_parser = subsubparser.add_parser('update', help='Update a recruiter')
    recruiter_modify_parser.add_argument('-b','--batch',type=str,required=True,help='batch in which workers are deployed')
    recruiter_modify_parser.add_argument('-n','--rank',type=int,help='rank in which recruiter is tried (unique per batch), default to 1',default=1)
    recruiter_modify_parser.add_argument('-f','--flavor',type=str,help='flavor (type of instance) of worker needed',default=None)
    recruiter_modify_parser.add_argument('-r','--region',type=str,help='worker region of provenance',default=None)
    recruiter_modify_parser.add_argument('-P','--provider',type=str,help='worker provider',default=None)
    recruiter_modify_parser.add_argument('-C','--concurrency',type=int,help='what should be the concurrency setting of recruited workers',default=None)
    recruiter_modify_parser.add_argument('-p','--prefetch',type=int,help='what should be the prefetch setting of recruited workers (default to 0)',default=None)
    recruiter_modify_parser.add_argument('-T','--tasks-per-worker',type=int,help='how many workers should be recruited (default to concurrency, but if you want each worker to work two rounds, set it to twice the concurrency)',default=None)
    recruiter_modify_parser.add_argument('-m','--minimum-tasks',type=int,help='prevent the recruiter from triggering before this minimal number of tasks is reached',default=None)
    recruiter_modify_parser.add_argument('-W','--maximum-workers',type=int,help='prevent the recruiter from recruiting more than this number of worker',default=None)

    recruiter_delete_parser = subsubparser.add_parser('delete', help='Delete a recruiter')
    recruiter_delete_parser.add_argument('-b','--batch',type=str,required=True,help='batch in which workers are deployed')
    recruiter_delete_parser.add_argument('-n','--rank',type=int,help='rank in which recruiter is tried (unique per batch), default to 1',default=1)
    

    db_parser = subparser.add_parser('db', help='The following options are to work with scitq database')
    db_parser.add_argument('-c','--conf', help=f'Use this environment file to set environment (default to {DEFAULT_SERVER_CONF})', type=str, default=DEFAULT_SERVER_CONF)
    subsubparser=db_parser.add_subparsers(dest='action')
    db_upgrade_parser=subsubparser.add_parser('upgrade',help='Migrate the database to the current version of scitq')
    db_init_parser=subsubparser.add_parser('init',help='Initialize a new database with current version of scitq')
    db_upgrade_parser=subsubparser.add_parser('upgrade-or-init',help='Migrate the database to the current version of scitq if it exists, initialize it otherwise')

    execution_parser = subparser.add_parser('execution', help='The following options will only concern task executions')
    subsubparser=execution_parser.add_subparsers(dest='action')
    list_execution_parser= subsubparser.add_parser('list',help='List all task executions')
    option_group=list_execution_parser.add_mutually_exclusive_group()
    option_group.add_argument('-i','--id',help='Filter for this task id',type=int, default=None)
    option_group.add_argument('-n','--name',help='Filter for this task name',type=str, default=None)
    list_execution_parser.add_argument('-S','--status',help='Filter by the executions statuses',type=str,choices=EXECUTION_STATUS,default=None)
    list_execution_parser.add_argument('-b','--batch',help='Give you a list of task executions according to his batch',type=str,default='')
    list_execution_parser.add_argument('-H','--no-header',help='Do not print the headers',action='store_true')
    list_execution_parser.add_argument('-l','--limit',help='Limit output to the N latest executions (default to 10)',type=int, default=10)
    list_execution_parser.add_argument('-L','--long',help=f'Print all available data (latest {MAX_LENGTH_STR} char for output/error)',action='store_true')

    output_execution_parser= subsubparser.add_parser('output',help='Show the output/error for an execution')
    option_group=output_execution_parser.add_mutually_exclusive_group()
    option_group.add_argument('-i','--id',help='Filter for this task id',type=int, default=None)
    option_group.add_argument('-n','--name',help='Filter for this task name',type=str, default=None)
    option_group.add_argument('-x','--execution-id',help='Filter for this task execution id',type=int, default=None)
    output_execution_parser.add_argument('-N','--relative-execution',help='0 means latest execution, -1 previous one, -2 the one before, etc.',type=int, default=None)
    output_execution_parser.add_argument('-l','--limit',help='Limit output to the N latest executions (default to 10)',type=int, default=10)
    option_group=output_execution_parser.add_mutually_exclusive_group()
    option_group.add_argument('-o','--output',help='Show only the output for a task',action='store_true')
    option_group.add_argument('-e','--error',help='Show only the error for a task',action='store_true')
    output_execution_parser.add_argument('-H','--no-header',help='Do not print the headers',action='store_true')
    output_execution_parser.add_argument('-S','--status',help='Filter by the executions statuses',type=str,choices=EXECUTION_STATUS,default=None)
    output_execution_parser.add_argument('-C','--columns',help='A comma separated list of columns',type=str,default=None)

    flavor_parser = subparser.add_parser('flavor', help='Get information on flavors (VM sizes)')
    subsubparser=flavor_parser.add_subparsers(dest='action')

    list_flavor_parser= subsubparser.add_parser('list',help='List available flavors')
    list_flavor_parser.add_argument('-P','--provider',type=str,help='restrict to this provider',default=None)
    list_flavor_parser.add_argument('-r','--region',type=str,help='restrict to this region',default=None)
    list_flavor_parser.add_argument('--min-cpu',type=int,help='restrict to VM with this minimum of CPU',default=0)
    list_flavor_parser.add_argument('--min-ram',type=int,help='restrict to VM with this minimum of RAM (Gb)',default=0)
    list_flavor_parser.add_argument('--min-disk',type=int,help='restrict to VM with this minimum of disk (Gb)',default=0)
    list_flavor_parser.add_argument('--max-eviction',type=int,
                                    help=f'restrict to a maximum of eviction (default: {FLAVOR_DEFAULT_EVICTION})',
                                    default=FLAVOR_DEFAULT_EVICTION)
    list_flavor_parser.add_argument('--limit',type=int,
                                    help=f'limit answer to that number of references (default: {FLAVOR_DEFAULT_LIMIT})',
                                    default=FLAVOR_DEFAULT_LIMIT)  
    list_flavor_parser.add_argument('--protofilters',type=str,
                                    help=f'Add some : separated filters like cpu>1 or tags#G - do not forget to quote as shell like to intrepret > signs...',
                                    default=None)    

    args=parser.parse_args()

    if args.version:
        print(f"Version: {package_version()}")
        return None

    s = Server(args.server, get_timeout=args.timeout)
    if args.object=='worker':
        if args.action =='list':
            info_worker=['worker_id','name','status','concurrency','creation_date','last_contact_date','batch']
            if args.long:
                info_worker+=['provider','region','flavor','ipv4','ansible-active','prefetch','assigned','accepted','running','failed','succeeded']
            if not args.no_header:
                headers = info_worker
            else:
                headers = []
            filters={}
            if args.batch:
                filters['batch']=args.batch
            if args.status:
                filters['status']=args.status
            worker_list=s.workers(**filters) 
            if args.long:
                worker_dict=dict([(w['worker_id'],w) for w in worker_list])
                worker_tasks=s.workers_tasks()
                for w in worker_tasks:
                    worker_dict[w['worker_id']][w['status']]=w['count']
            
            __list_print(worker_list, info_worker, headers)

        elif args.action =='deploy' :
            s.worker_deploy(number=args.number,batch=args.batch,region=args.region,
                flavor=args.flavor,concurrency=args.concurrency, prefetch=args.prefetch,
                provider=args.provider)

        elif args.action =='delete' :
            if args.id is not None:
                id=args.id
            elif args.name is not None:
                for worker in s.workers():
                    if worker['name']==args.name:
                        id=worker['worker_id']
                        break
                else:
                    raise RuntimeError(f'No such worker {args.name}...')
            else:
                raise RuntimeError('You must specify either name (-n) or id (-i)')
            s.worker_delete(id)

        elif args.action =='update' :
            if args.id is not None:
                id=args.id
            elif args.name is not None:
                for worker in s.workers():
                    if worker['name']==args.name:
                        id=worker['worker_id']
                        break
                else:
                    raise RuntimeError(f'No such worker {args.name}...')
            else:
                raise RuntimeError('You must specify either name (-n) or id (-i)')
            s.worker_update(id, batch=args.batch, status=args.status,
                concurrency=args.concurrency, prefetch=args.prefetch, flavor=args.flavor)
        
    elif args.object == 'batch':
        if args.action =='list':
            info_batch=['batch','pending','accepted','running','failed','succeeded','workers']
            batch_list=s.batches() 
            __list_print(batch_list, info_batch, info_batch)

        if args.action == 'stop':
            signal=0
            if args.number:
                signal = args.number
            elif args.term:
                signal = SIGQUIT
            elif args.kill:
                signal = SIGKILL
            elif args.pause:
                signal = SIGTSTP
            for batch in args.name :
                s.batch_stop(batch,signal)
        
        elif args.action == 'go':
            signal=0
            if args.number:
                signal = args.number
            elif args.cont:
                signal = SIGCONT
            for batch in args.name :
                s.batch_go(batch, signal)

        elif args.action == 'delete':
            for batch in args.name :
                s.batch_delete(batch)
        
        
    elif args.object == 'task':  
        if args.action == 'list':
            info_task=['task_id','name','status','command','creation_date','modification_date','status_date','batch']
            if args.long:
                info_task+=['container','container_options','input','output','resource','required_task_ids','retry','run_timeout','download_timeout']
            if not args.no_header:
                headers = info_task
            else:
                headers = []
            filters={}
            if args.batch:
                filters['batch']=args.batch
            if args.status:
                filters['status']=args.status
            task_list=s.tasks(**filters)
            __list_print(task_list, info_task, headers, long=args.long)

        elif args.action=='relaunch':
            if args.id is not None:
                id=args.id
            elif args.name is not None:
                for task in s.tasks(name=args.name):
                    id=task['task_id']
                    break
                else:
                    raise RuntimeError(f'No such task {args.name}...')       
            if s.task_get(id)['status']=='pending':
                print('This task is already in queue')
            else:
                s.task_update(id=id,status='pending')
        
        elif args.action == 'output':
            if args.id is not None:
                id=args.id
            elif args.name is not None:
                for task in s.tasks(name=args.name):
                    id=task['task_id']
                    break
                else:
                    raise RuntimeError(f'No such task {args.name}...')   
            executions = s.executions(task_id=id, latest=True)
            try:
                execution = next(iter(executions))
            except StopIteration:
                raise RuntimeError(f'No execution for this task {id}')
            if args.output:
                print(execution['output'])
            elif args.error:
                print(execution['error'])
            else:
                print(tabulate([[execution['output'],execution['error']]],headers=["output","error"]))
        
        elif args.action == 'update':
            if args.id is not None:
                id=args.id
            elif args.name is not None:
                for task in s.tasks(name=args.name):
                    id=task['task_id']
                    break
                else:
                    raise RuntimeError(f'No such task {args.name}...')  
            if args.requirements is not None:
                try:
                    args.requirements = [int(r) for r in args.requirements.split(' ')]
                except: 
                    try:
                        args.requirements = [int(r) for r in args.requirements.split(',')]
                    except:
                        raise RuntimeError(f'requirements should be a space or comma separated list of integer not {args.requirements}') 
            s.task_update(id, name=args.new_name, status=args.status, batch=args.batch,
                command=args.command, container=args.docker, container_options=args.option,
                input=args.input, output=args.output, required_task_ids=args.requirements, 
                run_timeout=args.run_timeout, download_timeout=args.download_timeout)
        elif args.action == 'delete':
            if args.id is not None:
                id=args.id
            elif args.name is not None:
                for task in s.tasks(name=args.name):
                    id=task['task_id']
                    break
                else:
                    raise RuntimeError(f'No such task {args.name}...')   
            else:
                raise RuntimeError('You must specify either name (-n) or id (-i)')
            s.task_delete(id)
    

    elif args.object=='ansible':

        if args.action=='install':
            if not os.path.exists(args.path):
                print('Creating directory', args.path)
                os.makedirs(args.path)
            if os.path.exists(os.path.join(args.path,'sqlite_inventory.py')):
                import hashlib
                with open(os.path.join(args.path,'sqlite_inventory.py'),'rb') as f:
                    script_md5 = hashlib.md5(f.read()).hexdigest()
                if script_md5==OLD_SQLITE_INVENTORY_MD5:
                    print(f'Removing old script',os.path.join(args.path,'sqlite_inventory.py'))
                    os.remove(os.path.join(args.path,'sqlite_inventory.py'))
                else:
                    print(f"WARNING: a file named {os.path.join(args.path,'sqlite_inventory.py')} was found, but it does not fit typical scitq inventory script, see if it should be removed manually.")
            print('Installing files')
            my_path,_ = os.path.split(os.path.realpath(sys.argv[0]))          
            shutil.copy(os.path.join(my_path, 'scitq-inventory'), 
                os.path.join(args.path,'scitq-inventory') )
            os.chmod(os.path.join(args.path,'scitq-inventory'), 0o770)
            shutil.copy(package_path('ansible','scitq','01-scitq-default'), 
                os.path.join(args.path,'01-scitq-default') )
        elif args.action=='path':
            print(package_path('ansible','playbooks'))
        elif args.action=='inventory':
            print('Deprecated use scitq-manage worker list -L instead')

    
    elif args.object=='debug':

        if args.action=='run':
            if args.id is not None:
                task = s.task_get(args.id)
            elif args.name is not None:
                for task in s.tasks():
                    if task['name']==args.name:
                        break
                else:
                    raise RuntimeError(f'No such task {args.name}...')       
            elif args.batch is not None:
                try:
                    task = random.choice([task for task in s.tasks() if task['batch']==args.batch and task['status']=='pending'])
                except IndexError:
                    raise RuntimeError(f'No pending task in this batch {args.batch}...')
            else:
                try:
                    task = random.choice([task for task in s.tasks() if task['status']=='pending'])
                except IndexError:
                    raise RuntimeError(f'No pending task...')
            get_input = get_resource = True
            if args.retry:
                get_input = get_resource = False
            if args.no_resource:
                get_resource = False
            Debugger(task, get_input=get_input, get_resource=get_resource, 
                     configuration=args.conf, extra_configuration=args.worker_conf).run()
    

    elif args.object=='recruiter':

        if args.action =='list':
            info_recruiter=['batch','rank','flavor','region','provider','concurrency']
            if args.long:
                info_recruiter+=['prefetch','tasks_per_worker','minimum_tasks','maximum_worker']
            if not args.no_header:
                headers = info_recruiter
            else:
                headers = []
            filters={}
            if args.batch:
                filters['batch']=args.batch
            __list_print(list([dict([(k.replace('worker_',''),v) for k,v in r.items()]) 
                     for r in s.recruiters(**filters)]), info_recruiter, headers)

        elif args.action =='create' :
            if args.tasks_per_worker is None:
                args.tasks_per_worker=args.concurrency
            s.recruiter_create(batch=args.batch,rank=args.rank,region=args.region,
                flavor=args.flavor,concurrency=args.concurrency, prefetch=args.prefetch,
                provider=args.provider, tasks_per_worker=args.tasks_per_worker,
                minimum_tasks=args.minimum_tasks, maximum_workers=args.maximum_workers)
        
        elif args.action == 'update':
            s.recruiter_update(batch=args.batch,rank=args.rank,region=args.region,
                flavor=args.flavor,concurrency=args.concurrency, prefetch=args.prefetch,
                provider=args.provider, tasks_per_worker=args.tasks_per_worker,
                minimum_tasks=args.minimum_tasks, maximum_workers=args.maximum_workers)

        elif args.action =='delete' :
            s.recruiter_delete(batch=args.batch, rank=args.rank)

    elif args.object == 'execution':

        if args.action == 'list':
            info_execution=['execution_id','task_id','worker_id','status','command','creation_date','modification_date','latest']

            if not args.no_header:
                headers = info_execution
            else:
                headers = []
            filters={}
            if args.long:
                filters['no_output']=False
                info_execution+=['output','error','output_files']
            else:
                filters['no_output']=True
            if args.batch:
                filters['batch']=args.batch
            if args.status:
                filters['status']=args.status
            if args.id:
                filters['task_id']=args.id
            if args.name:
                filters['task_name']=args.name
            if args.limit:
                filters['limit']=args.limit
            execution_list=s.executions(reverse=True, trunc=MAX_LENGTH_STR, **filters)
            __list_print(execution_list, info_execution, headers, long=False)
        
        if args.action == 'output':
            info_execution=['execution_id','task_id','worker_id','status','command','creation_date','modification_date','latest','output','error']
            if args.columns:
                new_info = []
                for c in args.columns.split(','):
                    c=c.strip()
                    if c not in info_execution:
                        raise RuntimeError(f'Column {c} is not available for executions, pick within {info_execution}')
                    new_info.append(c)
                info_execution = new_info
            if not args.no_header:
                headers = info_execution
            else:
                headers = []

            filters = {}
            if args.id:
                filters['task_id']=args.id
            if args.name:
                filters['task_name']=args.name
            if args.execution_id:
                filters['execution_id']=args.execution_id
            if args.limit and not args.relative_execution:
                filters['limit']=args.limit
            if args.status:
                filters['status']=args.status
            if args.output:
                info_execution.remove('error')
            if args.error:
                info_execution.remove('output')
            execution_list=s.executions(reverse=True, **filters)
            if args.relative_execution:
                initial_execution_list = execution_list
                execution_list = []
                current_task_id = None
                for e in initial_execution_list:
                    if e['task_id']!=current_task_id:
                        current_task_id=e['task_id']
                        n=0
                    if n==args.relative_execution:
                        execution_list.append(e)
                    n-=1
                    if args.limit and len(execution_list)==args.limit:
                        break
            __list_print(execution_list, info_execution, headers, long=True)


    elif args.object=='db':

        if args.action=='upgrade-or-init':
            dotenv.load_dotenv(args.conf)
            from .default_settings import SQLALCHEMY_DATABASE_URI
            from sqlalchemy import create_engine
            from sqlalchemy.exc import OperationalError
            engine = create_engine(SQLALCHEMY_DATABASE_URI)
            try:
                engine.connect()
                engine.execute('SELECT * FROM task LIMIT 1')
                args.action='upgrade'
            except OperationalError:
                args.action='init'

        if args.action=='init':
            dotenv.load_dotenv(args.conf)
            # return code of flask db init seems to be random, outputing to 1 with no reason or with a good reason...
            run('SCITQ_PRODUCTION=1 FLASK_APP=server flask db init', shell=True, 
                cwd=package_path(), check=False)
            run('SCITQ_PRODUCTION=1 FLASK_APP=server flask db stamp head', shell=True, 
                cwd=package_path(), check=True)
            print('DB intialized')

        if args.action=='upgrade':
            dotenv.load_dotenv(args.conf)
            run('SCITQ_PRODUCTION=1 FLASK_APP=server flask db upgrade', shell=True, 
                cwd=package_path(), check=True)
            print('DB migrated')

    elif args.object == 'flavor':

        if args.action=='list':
            flavors = s.flavors(min_cpu=args.min_cpu, min_ram=args.min_ram, min_disk=args.min_disk,
                        max_eviction=args.max_eviction, limit=args.limit, provider=args.provider,
                        region=args.region, protofilters=args.protofilters)
            headers = ['name','provider','region','cpu','ram','tags','gpu','gpumem','disk','cost','eviction','available']
            __list_print(flavors, headers, headers, long=True)
    


if __name__=="__main__":
    main()


import argparse
import os
import json
from time import sleep
import shutil
from .db import db
from .model import Worker, Flavor
from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import NoResultFound
import filelock
from .config import SCITQ_SERVER
from ..util import filter_none


# GLOBAL CONSTANTS
DEFAULT_GROUP = os.environ.get('ANSIBLE_DEFAULT_GROUP','Default')

# INTERNAL CONSTANTS
VERSION=2
# how many times we should try /etc/hosts
TIMEOUT_ETCHOSTS=30
ETCHOSTS_MARKER_LINE="# DO NOT CHANGE THIS LINE OR ANY LINE BELOW - SCITQ MANAGED ZONE"

# via https://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid-in-python
def check_pid(pid):        
    """ Check For the existence of a unix pid. """
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True

class Database:
    def __init__(self, app):
        """Connect to db"""
        with app.app_context():
            self.session = Session(db.engine)

    def list(self):
        """Main inventory function, list all hosts in all groups with _meta (hostvars)"""
        inventory = {}
        inventory['_meta'] = {}
        for worker,flavor in self.session.query(Worker,Flavor).join(Worker.flavor_detail,isouter=True).filter(Worker.ansible_active==True):
            if worker.ansible_group not in inventory:
                inventory[worker.ansible_group]=[]
            inventory[worker.ansible_group].append(worker.hostname)
            inventory['_meta'][worker.hostname]=filter_none({
                'ipv4': worker.ipv4,
                'ipv6': worker.ipv6,
                'provider': worker.provider,
                'region': worker.region,
                'flavor': worker.flavor,
                'cpu': flavor.cpu if flavor else None,
                'ram': flavor.ram if flavor else None,
                'gpu': flavor.gpu if flavor else None,
                'target': SCITQ_SERVER,
                'tags': flavor.tags if flavor else None,
            })
        return json.dumps(inventory,  indent=4)

    def list_host(self, host_name):
        """List all hostvars for a certain host, return an empty json if host has no vars or does not exists"""
        try:
            worker,flavor = self.session.query(Worker,Flavor).join(Worker.flavor_detail,isouter=True).filter(Worker.ansible_active==True).filter(Worker.hostname==host_name).one()
            return json.dumps(filter_none({
                    'ipv4': worker.ipv4,
                    'ipv6': worker.ipv6,
                    'provider': worker.provider,
                    'region': worker.region,
                    'flavor': worker.flavor,
                    'cpu': flavor.cpu if flavor else None,
                    'ram': flavor.ram if flavor else None,
                    'gpu': flavor.gpu if flavor else None,
                    'target': SCITQ_SERVER,
                    'tags': flavor.tags if flavor else None,
                    }),
                    indent=4)
        except NoResultFound:
            return json.dumps({})

    def add_host(self, host_name, group_name):
        """Add a host in the inventory with default group_name - watch out: must
        be committed with self.session.commit()"""
        worker = self.session.query(Worker).filter(Worker.hostname==host_name).one()
        worker.ansible_group = group_name
        worker.ansible_active = True
            

    def set_hostvar(self, host_name, variable, value):
        """Insert a new host variable (or update) in table hostvars - value is converted to text"""
        worker = self.session.query(Worker).filter(Worker.hostname==host_name).one()
        if hasattr(worker, variable):
            setattr(worker, variable, value)
        else:
            print(f'WARNING: variable {variable} cannot be set, it is read-only, ignoring set to {value}')
    
    def del_host(self, host_name):
        """Delete a host and its hostvars in database"""
        worker = self.session.query(Worker).filter(Worker.hostname==host_name).one()
        worker.ansible_active = False
        
    def get_host_ips(self):
        """Return a dictionnary of hostname:IPv4"""
        return dict(self.session.query(Worker).filter(Worker.ansible_active==True).with_entities(Worker.hostname, Worker.ipv4))
    
    def commit(self):
        """Commit pending changes"""
        self.session.commit()

def decorate_parser(parser):
    """Easing integration: all options to parser added here"""
    parser.add_argument('--list', default=None, action='store_true',
                help="List all hosts")
    parser.add_argument('--host', type=str, action='store', default=None,
                help="List one host in particular")
    parser.add_argument('--add-host', type=str, action='store', default=None,
                help=f"Add a host (in group {DEFAULT_GROUP} unless --in-group is used)")
    parser.add_argument('--in-group', type=str, action='store', default=DEFAULT_GROUP,
                help=f"Specify in which group a host should be added")
    parser.add_argument('--for-host', type=str, default=None,
                help="Specify a host where to add variables")
    parser.add_argument('--variable', type=str, action='append',
                help="Add a host variable (requires the host to be known, either with --add-host or with --for-host, and requires a value set with --value), can be used several times")
    parser.add_argument('--value', type=str, action='append',
                help="Add a host variable value (requires --variable, and thus the host to be known - with --add-host or --for-host), can be used several times")
    parser.add_argument('--del-host', type=str, action='store', default=None,
                help=f"Delete a host")
    parser.add_argument('--change-etchosts', action='store_true',
                help="Modify /etc/hosts so that declared hosts in database are present in /etc/hosts (as ansible is very bad at this task)")


def change_etc_hosts(db):
    """Change /etc/hosts
    """
    try:
        with filelock.FileLock('/etc/hosts.lock', timeout=TIMEOUT_ETCHOSTS):
            host_ips = db.get_host_ips()
            new_hosts = list(host_ips.keys())
            shutil.copy('/etc/hosts','/etc/hosts.back')
            with open('/etc/hosts','r', encoding='utf-8') as f:
                etchost  = f.read().strip().split('\n')
            found_marker = False
            new_etchost = []
            for line in etchost:
                if not found_marker:
                    new_etchost.append(line)
                    if ETCHOSTS_MARKER_LINE in line:
                        found_marker=True
                        continue
                    else:
                        continue
                _,host=line.split()
                if host in new_hosts:
                    new_hosts.remove(host)
                    new_etchost.append(f"{host_ips[host]}\t{host}")
            if not found_marker:
                new_etchost.append(ETCHOSTS_MARKER_LINE)
            for host in new_hosts:
                new_etchost.append(f"{host_ips[host]}\t{host}")
            with open('/etc/hosts','w', encoding='utf-8') as f:
                f.write('\n'.join(new_etchost))
    except filelock.Timeout:
        pass

            


def inventory(app):
    """Main fonction"""
    parser = argparse.ArgumentParser(description='Ansible SQLite inventory script')
    decorate_parser(parser)
    args = parser.parse_args()
    
    if args.variable or args.value:
        if len(args.variable)!=len(args.value):
            raise (f"Use exactly the same number of values (value used: {args.value}) and variables (variables used: {args.variable})")
        if args.add_host is None and args.for_host is None:
            raise ("Cannot set some variables with values if host is unknown, use --for-host or --add-host")
        variables = dict(zip(args.variable, args.value))
    else:
        variables = {}

    db = Database(app)

    if args.list:
        return db.list()
    elif args.host is not None:
        return db.list_host(args.host)
    
    if args.add_host is not None:
        host = args.add_host
        db.add_host(host, args.in_group)
        if 'ansible_host' not in variables:
            variables['ansible_host']=host
    elif args.for_host is not None:
        host = args.for_host
    
    if variables:
        for variable, value in variables.items():
            db.set_hostvar(host, variable, value)

    if args.del_host is not None:
        db.del_host(args.del_host)
    
    db.commit()

    if args.change_etchosts:
        change_etc_hosts(db)
    



from threading import Thread
import os
import logging as log
from .config import WORKER_CREATE, setup_log
from flask import Flask
from werkzeug.middleware.proxy_fix import ProxyFix
from flask_migrate import Migrate
from sqlalchemy.orm import Session

from .db import db

def __background__(*args):
    raise RuntimeError('Not initialized')

migrate=Migrate()

def create_app(get_background=True, get_webapp=True):
    setup_log()
    log.info('Starting')
    
    # via https://github.com/pallets/flask-sqlalchemy/blob/main/examples/hello/hello.py
    app = Flask(__name__, instance_relative_config=True)
    app.config.from_object('scitq.default_settings')

    # via https://flask-restx.readthedocs.io/en/latest/example.html
    app.wsgi_app = ProxyFix(app.wsgi_app)
    
    from . import model
    db.init_app(app)

    # with uwsgi, the master worker is forking to create the workers which receive a 
    # non-working connection (because it comes from another process), it must be discarded
    # so that workers re-open the connection properly
    # idea from https://stackoverflow.com/questions/39562838/how-to-configure-pyramid-uwsgi-sqlalchemy
    try:
        # import uwsgi is only working in uwsgi context. It is normal that is fails
        # to import in VisualStudioCode or manually
        import uwsgi # pyright: ignore[reportMissingImports]

        def postfork():
            with app.app_context():
                db.engine.dispose()
        uwsgi.post_fork_hook = postfork
    except ImportError:
        pass
    
    with app.app_context():
        db.create_all()
    migrate.init_app(app, db)

    from .api import api
    api.init_app(app)

    from .ui import ui
    app.register_blueprint(ui)

    if get_background:
        from .background import background
        global __background__
        __background__=background
    


    if get_webapp and not os.environ.get('SCITQ_PRODUCTION'):
        Thread(target=background, args=[app]).start()

    return app


def background_app():
    """This is run by scitq-queue service"""
    app=create_app()
    __background__(app)

def ansible_inventory():
    """This is run when calling scitq-inventory"""
    app=create_app(get_background=False, get_webapp=False)
    from .inventory import inventory
    result = inventory(app)
    if result:
        print(result)

def get_session():
    """Return a simple session for the database access"""
    app=create_app(get_background=False, get_webapp=False)
    with app.app_context():
        session = Session(db.engine)
    return session

if os.environ.get('SCITQ_PRODUCTION'):
    app = create_app()


SIGNAL_CLEAN=-1
SIGNAL_RESTART=-2
SIGNAL_RESET_RESOURCES=-3

DEFAULT_SERVER_CONF='/etc/scitq.conf'
DEFAULT_WORKER_CONF='/etc/scitq-worker.conf'

HTTP_ERROR_CODE_NOT_FOUND = 404

TASK_STATUS = ['paused','waiting','pending','assigned','accepted','running','failed','succeeded']
EXECUTION_STATUS = ['pending','accepted','running','failed','succeeded','refused']
WORKER_STATUS = ['paused','running','offline','failed','evicted']

PROTOFILTER_SYNTAX=r'^(?P<item>cpu|ram|disk|tags|gpumem|eviction|region|cost|provider)(?P<comparator>==|!=|>=|>|<|<=|~|#|!~|!#)(?P<value>[0-9\.]+|[A-Za-z0-9_%-]+)$'
PROTOFILTER_SEPARATOR=':'

FLAVOR_DEFAULT_LIMIT=10
FLAVOR_DEFAULT_EVICTION=5

TASK_STATUS_ID = {status:i for i,status in enumerate(TASK_STATUS)}
TASK_STATUS_ID_REVERSE = {i:status for i,status in enumerate(TASK_STATUS)}


import threading
import multiprocessing
import traceback
import os
import boto3
from configparser import ConfigParser
import re
import stat
import shutil
from argparse import Namespace
from .constants import PROTOFILTER_SYNTAX, PROTOFILTER_SEPARATOR
from functools import reduce
import hashlib
import sys

MAX_MD5_BLOCK_SIZE=1024**2

class PropagatingThread(threading.Thread):
    """Taken from https://stackoverflow.com/questions/2829329/catch-a-threads-exception-in-the-caller-thread
    the join here will fails if one of the thread fails.
    """
    def run(self):
        """This version of run capture the exception occuring in the thread"""
        self.exc = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self, timeout=None):
        """This version of join throw an exception if the thread joined did
        throw an exception"""
        super(PropagatingThread, self).join(timeout)
        if self.exc:
            raise self.exc
        return self.ret

    def is_alive(self):
        """This version of is_alive throw an exception if the thread ended
        with an exception"""
        if self.exc:
            raise self.exc
        return super(PropagatingThread, self).is_alive()

class PropagatingProcess(multiprocessing.Process):
    """Same as above but for a process: this is a process that propagate exceptions.
    """
    def __init__(self, *args, **kwargs):
        super(PropagatingProcess, self).__init__(*args, **kwargs)
        self._pconn, self._cconn = multiprocessing.Pipe()
        self._exception = None

    def run(self):
        try:
            super(PropagatingProcess, self).run()
            self._cconn.send(None)
        except Exception as e:
            tb = traceback.format_exc()
            self._cconn.send((e, tb))

    @property
    def exception(self):
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        return self._exception

    def join(self, timeout=None):
        """This version of join throw an exception if the process joined did
        throw an exception"""
        ret=super(PropagatingProcess, self).join(timeout)
        if self.exception:
            error, tb = self.exception
            print(tb)
            raise error
        return ret

    def is_alive(self):
        """This version of is_alive throw an exception if the process ended
        with an exception"""
        if self.exception:
            error, tb = self.exception
            print(tb)
            raise error
        return super(PropagatingProcess, self).is_alive()

def package_path(*subdirs):
    """A very stupid hack to point to actual package data"""
    return os.path.join(os.path.dirname(__file__), *subdirs)

def package_version():
    """Return scitq package version"""
    import pkg_resources
    return pkg_resources.get_distribution(__package__).version


class xboto3:
    # this is an ugly hack to bypass the fact boto3 does not read all info
    # in .aws/config, notably OVH specific options, endpoint_url
    """This class is a replacement of boto3 library. It should be used like this:
    instead of:
    
    boto3.resource('s3') -> xboto3().resource('s3')

    boto3.client('s3') -> xboto3().client('s3')

    The constructor (xboto3()) accept a profile_name argument.

    """

    BOTO3_CONFIG_REGEXP = re.compile('(\S+) = (".*?"|\S+)')
    BOTO3_ACCEPTED_OPTIONS = ['endpoint_url']
    
    def __init__(self, profile_name='default'):
        """Retrieves boto3.resource, and fills in any service-specific 
        (filtered by BOTO3_ACCEPTED_OPTIONS, as some options are not used)    
        parameters from your config, which can be specified either        
        in AWS_CONFIG_FILE or ~/.aws/config (default). Similarly,         
        profile_name is 'default' (default) unless AWS_PROFILE is set.    
                                                                        
        Assumes that additional service-specific config is specified as:  
                                                                        
        [profile_name]                                                    
        service-name =                                                    
            parameter-name = parameter-value                              
        another-service-name =                                            
            ... etc                                                       
        
        adapted from: https://github.com/aws/aws-cli/issues/1270
        thanks to https://github.com/jaklinger
        """
        # Get the AWS config file path                                    
        self.profile_name = os.environ.get('AWS_PROFILE', profile_name)
        self.conf_filepath = os.environ.get('AWS_CONFIG_FILE', '~/.aws/config')
        self.aws_endpoint_url = os.environ.get("AWS_ENDPOINT_URL")
        self.conf_filepath = os.path.expanduser(self.conf_filepath)
        
        self.service_cfg = {}
        
        if self.aws_endpoint_url:
            # Use environment variable if available
            self.service_cfg['endpoint_url']=self.aws_endpoint_url

    def __get_config__(self, service_name):
        config = dict(self.service_cfg)
        if os.path.exists(self.conf_filepath):
            parser = ConfigParser()
            with open(self.conf_filepath) as f:
                parser.read_file(f)
            cfg = dict(parser).get(f'profile {self.profile_name}', {})
            # Extract the service-specific config, if any                     
            service_raw_cfg = cfg.get(service_name, '')
            config.update({k: v for k, v in self.BOTO3_CONFIG_REGEXP.findall(service_raw_cfg)
                if k in self.BOTO3_ACCEPTED_OPTIONS})
        return config
        
    def resource(self, service_name, **kwargs):
        # Load in the service config, on top of other defaults            
        # and let boto3 do the rest                                       
        return boto3.resource(service_name=service_name,
                                **self.__get_config__(service_name), **kwargs)
    

    def client(self, service_name, **kwargs):
        # Load in the service config, on top of other defaults            
        # and let boto3 do the rest                                       
        return boto3.Session().client(
            service_name=service_name, **self.__get_config__(service_name), **kwargs
        )
    

def check_dir(path):
    """Try to create dir for this path if it does not exists"""
    path,_ = os.path.split(path)
    if not os.path.exists(path):
        os.makedirs(path)

def to_dict(obj):
    """Return a dict from an object filtering on "normal" attributes"""
    return {k:v if type(v) in [str,int,float] else '' if v is None else str(v) for k,v in obj.__dict__.items() if not k.startswith('_')}

def tryupdate(mydict, key, f, *args, alternative=None, **kwargs):
    """Try to compute a certain value with a certain function and update a dict with it"""
    try:
        value=f(*args, **kwargs)
        mydict[key]=value
    except:
        if alternative is not None:
            mydict[key]=alternative
    return mydict

def isfifo(filepath):
    """Return True if file is a fifo"""
    return stat.S_ISFIFO(os.stat(filepath).st_mode)

def if_is_not_None(x, default):
    """A small helper function that replace None with a default value"""
    return x if x is not None else default

class colors:
    '''Colors class:reset all colors with colors.reset; two
    sub classes fg for foreground
    and bg for background; use as colors.subclass.colorname.
    i.e. colors.fg.red or colors.bg.greenalso, the generic bold, disable,
    underline, reverse, strike through,
    and invisible work with the main class i.e. colors.bold'''
    reset = '\033[0m'
    bold = '\033[01m'
    disable = '\033[02m'
    underline = '\033[04m'
    reverse = '\033[07m'
    strikethrough = '\033[09m'
    invisible = '\033[08m'
    blinking = '\033[05m'
    
    class fg:
        black = '\033[30m'
        red = '\033[31m'
        green = '\033[32m'
        orange = '\033[33m'
        blue = '\033[34m'
        purple = '\033[35m'
        cyan = '\033[36m'
        lightgrey = '\033[37m'
        darkgrey = '\033[90m'
        lightred = '\033[91m'
        lightgreen = '\033[92m'
        yellow = '\033[93m'
        lightblue = '\033[94m'
        pink = '\033[95m'
        lightcyan = '\033[96m'
        white = '\033[97m'
 
    class bg:
        black = '\033[40m'
        red = '\033[41m'
        green = '\033[42m'
        orange = '\033[43m'
        blue = '\033[44m'
        purple = '\033[45m'
        cyan = '\033[46m'
        lightgrey = '\033[47m'

def flat_list(l):
    """Return a list of value from a list of one-uplet containing the value"""
    return list([item[0] for item in l])

def force_hard_link(src, dst):
    """A function that force hardlink creation, possibly overriding anything existing before"""
    if os.path.exists(dst):
        if os.path.isdir(dst):
            shutil.rmtree(dst)
        else:
            os.remove(dst)
    os.link(src, dst)

def filter_none(d):
    """filter out null value from a dict as well as key in remove list (default
    to ['id'])"""
    return dict([(k,v) for k,v in d.items() if v is not None])

def to_obj(d):
    """Transform a dict into an obj"""
    return Namespace(**d)

protofilter_syntax=re.compile(PROTOFILTER_SYNTAX)

def validate_protofilter(worker_flavor):
    """A function that raise an Exception if worker_flavor does not match expected syntax"""
    if worker_flavor.startswith('auto'):
        if PROTOFILTER_SEPARATOR in worker_flavor:
            for filter in worker_flavor.split(PROTOFILTER_SEPARATOR)[1:]:
                match=protofilter_syntax.match(filter)
                if not match:
                    raise RuntimeError(f'Protofilter syntax error in {worker_flavor}: {filter} is not recognized as a proper filter')
        else:
            if worker_flavor!='auto':
                raise RuntimeError(f'Protofilter syntax error, worker_flavor should be auto or auto{PROTOFILTER_SEPARATOR}... not {worker_flavor}')

def is_like(a,b):
    """A small helper to simulate SQL like function that use % as a widechar"""
    return bool(re.match(b.replace('%','.*'),a))

def has_tag(a,b):
    """A small helper to simulate protofilter # sign, e.g. a#b == a contain tags b"""
    return reduce(lambda x,y: x and y, [c in a for c in b], True)

def bytes_to_hex(byte_array):
    """Convert a byte array to hex"""
    return None if byte_array is None else byte_array.hex()

def get_md5(path):
    """Return the md5 of a local file"""
    with open(path, "rb") as f:
        if sys.version_info>=(3,11):
            return hashlib.file_digest(f, "md5").hexdigest()
        else:
            md5=hashlib.md5()
            while True:
                chunk=f.read(MAX_MD5_BLOCK_SIZE)
                if not chunk:
                    break
                md5.update(chunk)
            return md5.hexdigest()


def split_list(l, n):
    """Split list l in n parts of approximately the same size"""
    # avoid creating empty sublist
    n = min(n, len(l))
    k, m = divmod(len(l), n)
    return (l[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(n))


from flask_restx import Api, Resource, fields
from flask import jsonify
from datetime import datetime
from sqlalchemy import and_, delete, select, func
from sqlalchemy.sql.expression import label
import logging as log
import json as json_module

from .model import Task, Execution, Signal, Requirement, Worker,\
    create_worker_destroy_job, Job, Recruiter, delete_batch, \
    find_flavor, execution_update_status, worker_delete, \
    ModelException, create_worker_create_job, worker_handle_eviction
from .db import db
from .config import IS_SQLITE
from ..constants import TASK_STATUS, EXECUTION_STATUS, FLAVOR_DEFAULT_LIMIT, FLAVOR_DEFAULT_EVICTION, WORKER_STATUS, TASK_STATUS_ID


api = Api(version='1.2', title='TaskMVC API',
    description='A simple TaskMVC API'
)



def process_filtering_args(ObjectType, args):
    """In list functions it seems natural to filter the results with some attributes
    conditions (such as 'batch="foo"' or 'status="succeeded"') and it makes more 
    efficient queries than listing the lot and filtering the results out or sending
    several individual queries (one per item id).
    However sqlalchemy filtering syntax is not friendly when using typical python
    optional args dictionary, this function does this transformation:
    
    When args maybe something as 'batch="foo"' or 'batch=["foo","bar"]'
    - the function will return 'ObjectType.batch=="foo"'
    - or 'ObjectType.batch.in_(["foo","bar"])' if value is a list

    which can then be passed to sqlalchemy .filter() function
    """
    # 
    # filter expects something as 'ObjectType.batch=="foo"'
    # or 'ObjectType.batch.in_(["foo","bar"])' if value is a list
    new_args = [ getattr(ObjectType,attribute).in_(value) if type(value)==list else \
                 getattr(ObjectType,attribute)==value 
                    for attribute,value in args.items() ]
    return and_(*new_args) if len(new_args)>1 else new_args[0]



class BaseDAO(object):
    ObjectType = None
    authorized_status = []

    def get(self, id):
        object = self.ObjectType.query.get(id)
        if object is None:
            api.abort(404, "{} {} doesn't exist".format(
                    self.ObjectType.__name__,id))
        return object
        
    def create(self, data):
        object = self.ObjectType(**data)
        if 'status' in data and data['status'] not in self.authorized_status:
            api.abort(500,
                f"Status {data['status']} is not possible (only {' '.join(self.authorized_status)})")
        db.session.add(object)
        db.session.commit()
        return object

    def update(self, id, data):
        object = self.get(id)
        modified = False
        for attr, value in data.items():
            if hasattr(object,attr): 
                if getattr(object,attr)!=value:
                    if attr=='status' and value not in self.authorized_status:
                        api.abort(500,
                           f"Status {value} is not possible (only {' '.join(self.authorized_status)})")
                    setattr(object, attr, value)
                    modified = True
            else:
                api.abort(500,f'Error: {object.__name__} has no attribute {attr}')
        if modified:
            if hasattr(object,'modification_date'):
                object.modification_date = datetime.utcnow()
            db.session.commit()
        return object

    def delete(self, id):
        object = self.get(id)
        db.session.delete(object)
        db.session.commit()
        return object
    
    def list(self, sorting_column=None, **args):
        if args:
            final_filter = process_filtering_args(self.ObjectType, args)
            return list(self.ObjectType.query.filter(final_filter).order_by(sorting_column))
        else:
            return list(self.ObjectType.query.order_by(sorting_column).all())
        
        



ns = api.namespace('tasks', description='TASK operations')

class TaskDAO(BaseDAO):
    ObjectType = Task
    authorized_status = TASK_STATUS

    def list(self, **args):
        task_list = []
        for task in super().list(sorting_column='task_id',**args):
            task.required_task_ids = list([r.other_task_id for r in task.requirements])
            task_list.append(task)
        return task_list

    def update(self, id, data):
        task = self.get(id)
        modified = False
        status_changed = False
        for attr, value in data.items():
            if hasattr(task,attr): 
                if getattr(task,attr)!=value:
                    if attr=='status':
                        if value not in self.authorized_status:
                            api.abort(500,
                                f"Status {value} is not possible (only {' '.join(self.authorized_status)})")
                        if task.status=='running':
                            for execution in db.session.query(Execution).filter(
                                    Execution.task_id==id, Execution.status=='running'):
                                db.session.add(Signal(execution.execution_id, execution.worker_id, 9))
                                execution.status='failed'
                                execution.modification_date = datetime.utcnow()
                        if value=='failed':
                            if 'retry' in data:
                                if data['retry']>0:
                                    value='pending'
                                    data['retry'] -= 1
                            elif task.retry>0:
                                value='pending'
                                task.retry -= 1
                        status_changed = True
                        
                    setattr(task, attr, value)
                    modified = True
            else:
                raise Exception('Error: {} has no attribute {}'.format(
                    task.__name__, attr))
        if modified:
            task.modification_date = datetime.utcnow()
            if status_changed:
                task.status_date = datetime.utcnow()
            db.session.commit()
        return task



task_dao = TaskDAO()

class RequirementDAO(BaseDAO):
    ObjectType = Requirement

    def list(self, **args):
        super().list(sorting_column='requirement_id', **args)

requirement_dao = RequirementDAO()


task = api.model('Task', {
    'task_id': fields.Integer(readonly=True, description='The task unique identifier'),
    'name': fields.String(required=False, description='The task (optional) name'),
    'status': fields.String(required=False, 
        description=f'The task status: {", ".join(TaskDAO.authorized_status)}'), 
    'command': fields.String(required=True, description='The task command'), 
    'creation_date': fields.DateTime(readonly=True, 
        description='timestamp of task creation in server'),
    'modification_date': fields.DateTime(readonly=True,
        description='timestamp of task last modification'),
    'status_date': fields.DateTime(readonly=True,
        description='timestamp of task last change of status'),
    'batch': fields.String(required=False, 
        description="only workers with the same batch (null or not) will accept the task."),
    'input': fields.String(required=False,
        decription="Input data required for task (space separated files URL in s3://...)"),
    'output': fields.String(required=False,
        decription="Output data basepath where /output content will be copied (if null, the result transmission is up to the command)"),
    'container': fields.String(required=False,
        decription="Container (as for now this is a docker) in which task is launched"),
    'container_options': fields.String(required=False,
        decription="Container (extra) option if needed"),
    'resource': fields.String(required=False,
        decription="Resource data required for task (much like input except it is shared between tasks) (space separated files URL in s3://...)"),
    'required_task_ids': fields.List(fields.Integer,required=False,
        description="List of task ids required to do this task"),
    'retry': fields.Integer(required=False,
        description="If set, retry the task this number of time if it fails until it succeeds"),
    'download_timeout': fields.Integer(required=False,
        description="If set, the task will timeout and fail if the download time exceeds this number in seconds"),
    'run_timeout': fields.Integer(required=False,
        description="If set, the task will timeout and fail if the run time exceeds this number in seconds"),
})

task_filter = api.model('TaskFilter', {
    'task_id': fields.List(fields.Integer(),required=False,decription='A list of ids to restrict listing'),
    'batch': fields.String(required=False, description="Filter with this batch"),
    'status': fields.String(required=False, description="Filter with this status"),
    'name': fields.String(required=False, description="Filter with this name"),
})


@ns.route('/')
class TaskList(Resource):
    '''Shows a list of all tasks, and lets you POST to add new tasks'''
    @ns.doc('list_tasks')
    @ns.expect(task_filter)
    @ns.marshal_list_with(task)
    def get(self):
        '''List all tasks'''
        return task_dao.list(**api.payload) if api.payload else task_dao.list()

    @ns.doc('create_task')
    @ns.expect(task)
    @ns.marshal_with(task, code=201)
    def post(self):
        '''Create a new task'''
        if 'required_task_ids' in api.payload:
            requirements = api.payload['required_task_ids']
            del(api.payload['required_task_ids'])
        else:
            requirements = []
        task = task_dao.create(api.payload)
        for r in requirements:
            requirement_dao.create({'task_id':task.task_id, 'other_task_id':r})
        return task, 201


task_status_filter = api.model('TaskStatusFilter', {
    'task_id': fields.List(fields.Integer(),required=True,decription='A list of ids to restrict listing'),
})

@ns.route('/status')
class TaskStatusList(Resource):
    '''Shows a list of all tasks, and lets you POST to add new tasks'''
    @ns.doc('list_task_statuses')
    @ns.expect(task_status_filter)
    def get(self):
        '''List all task status'''
        task_id_list = api.payload['task_id']
        task_status = db.session.query(Task.task_id,Task.status).filter(Task.task_id.in_(task_id_list)).all()
        task_to_status = { task_id:TASK_STATUS_ID[status] for task_id,status in task_status }
        return jsonify([task_to_status[id] for id in task_id_list])



@ns.route("/<id>")
@ns.param("id", "The task identifier")
@ns.response(404, "Task not found")
class WorkerObject(Resource):
    @ns.doc("get_task")
    @ns.marshal_with(task)
    def get(self, id):
        """Fetch a task given its identifier"""
        return task_dao.get(id)

    @ns.doc("update_task")
    @ns.expect(task)
    @ns.marshal_with(task, code=201)
    def put(self, id):
        """Update a task"""
        if 'required_task_ids' in api.payload:
            requirements = api.payload['required_task_ids']
            del(api.payload['required_task_ids'])
            db.session.execute(delete(Requirement).where(Requirement.task_id==id))
            for r in requirements:
                requirement_dao.create({'task_id':id, 'other_task_id':r})
        return task_dao.update(id, api.payload)

    @ns.doc("delete_task")
    @ns.marshal_with(task)
    def delete(self, id):
        """Delete a task"""
        return task_dao.delete(id)

ns = api.namespace('workers', description='WORKER operations')



class WorkerDAO(BaseDAO):
    ObjectType = Worker
    authorized_status = WORKER_STATUS

    def update_contact(self, id, load,memory,stats):
        db.engine.execute(
            db.update(Worker
                    ).values({'last_contact_date':datetime.utcnow(),
                        'load':load,'memory':memory,'stats':stats}
                    ).where(Worker.worker_id==id)
        )
        db.session.commit()

    def update(self, id, data):
        object = self.get(id)
        modified = False
        for attr, value in data.items():
            if hasattr(object,attr): 
                if getattr(object,attr)!=value:
                    if attr=='status':
                        if value not in self.authorized_status:
                            api.abort(500,
                                f"Status {value} is not possible (only {' '.join(self.authorized_status)})")
                        if object.status=='evicted' and value not in ['running']:
                            log.warning(f'Evicted worker {object.worker_id} can only return to running')
                            continue
                    setattr(object, attr, value)
                    if attr=='status' and value=='evicted':
                        worker_handle_eviction(worker=object,session=db.session, commit=False)
                    modified = True
            else:
                api.abort(500,f'Error: {object.__name__} has no attribute {attr}')
        if modified:
            if hasattr(object,'modification_date'):
                object.modification_date = datetime.utcnow()
            db.session.commit()
        return object

    def delete(self,id,is_destroyed=False, session=db.session):
        """Delete a worker
        """
        worker=self.get(id)
        log.warning(f'Deleting worker {id} ({worker.permanent})')
        return worker_delete(worker, session, is_destroyed=is_destroyed)

    def list(self, **args):
        return super().list(sorting_column='worker_id', **args)

worker_dao = WorkerDAO()

worker = api.model('Worker', {
    'worker_id': fields.Integer(readonly=True, description='The worker unique identifier'),
    'name': fields.String(required=True, description='The worker name'),
    'concurrency': fields.Integer(required=True, 
        description='The worker concurrency (nb of parallel processes)'), 
    'prefetch': fields.Integer(required=False, 
        description='How many jobs should be prefetch so that they can be launched as soon as possible'),     
    'hostname': fields.String(required=False, description='The worker hostname'),
    'status': fields.String(required=False,
         description=f'The worker status: {", ".join(WorkerDAO.authorized_status)}'), 
    'load': fields.String(readonly=True, description='The worker load (in %)'),
    'memory':fields.Float(readonly=True, description='Memory used (in %)'),
    'stats':fields.String(readonly=True,description='Other worker stats'),
    'task_properties':fields.String(required=False,description='Some task properties for batch transitions'),
    'creation_date': fields.DateTime(readonly=True, 
        description='timestamp of worker creation'),
    'modification_date': fields.DateTime(readonly=True, 
        description='timestamp of last worker modification'),
    'last_contact_date': fields.DateTime(readonly=True, 
        description='timestamp of last worker ping (automatically sent by worker'),
    'batch': fields.String(required=False, 
        description="worker accept only tasks with same batch (null or not)."),
    'permanent': fields.Boolean(readonly=False, required=False,
        description="Set to True to create a permanent worker"),
    'flavor': fields.String(required=False, 
        description="flavor (cloud type of instance) of the worker."),
    'region': fields.String(readonly=True, 
        description="region (cloud regional entity of the instance) of the worker."),
    'provider': fields.String(readonly=True, 
        description="provider (cloud provider of the instance) of the worker."),
})

@ns.route('/')
class WorkerList(Resource):
    '''Shows a list of all workers, and lets you POST to add new workers'''
    @ns.doc('list_workers')
    @ns.marshal_list_with(worker)
    def get(self):
        '''List all workers'''
        return worker_dao.list(**api.payload)

    @ns.doc('create_worker')
    @ns.expect(worker)
    @ns.marshal_with(worker, code=201)
    def post(self):
        '''Create a new worker'''
        return worker_dao.create(api.payload), 201


worker_tasks = api.model('WorkerTasks', {
    'worker_id': fields.Integer(readonly=True, description='A worker unique identifier'),
    'status': fields.String(readonly=True, description='The different status of tasks'), 
    'count': fields.Integer(readonly=True, description='How many tasks of this type'),
})

@ns.route("/tasks")
class WorkerTaskList(Resource):
    '''Shows a list of all tasks in all workers, and lets you POST to add new workers'''
    @ns.doc('list_worker_tasks')
    @ns.marshal_list_with(worker_tasks)
    def get(self):
        '''List all workers'''
        return list(db.session.execute("""SELECT w.worker_id, e.status, count(e.task_id) as count 
            FROM worker w
            JOIN execution e ON (e.worker_id=w.worker_id)
            JOIN task t ON (e.task_id=t.task_id 
                AND e.latest)
            GROUP BY w.worker_id,e.status"""))
    

@ns.route("/<id>")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerObject(Resource):
    @ns.doc("get_worker")
    @ns.marshal_with(worker)
    def get(self, id):
        """Fetch a worker given its identifier"""
        return worker_dao.get(id)

    @ns.doc("update_worker")
    @ns.expect(worker)
    @ns.marshal_with(worker, code=201)
    def put(self, id):
        """Update a worker"""
        #worker_dao.update_contact(id)
        if 'batch' in api.payload and api.payload['batch']=='':
            api.payload['batch']==None
        return worker_dao.update(id, api.payload)

    @ns.doc("delete_worker")
    @ns.marshal_with(worker)
    def delete(self, id):
        """Delete a worker"""
        return worker_dao.delete(id)

ping_parser = api.parser()
ping_parser.add_argument('load', type=str, help='Worker load', location='json')
ping_parser.add_argument('memory', type=float, help='Worker memory', location='json')
ping_parser.add_argument('stats', type=str, help='Worker other stats', location='json')

@ns.route("/<id>/ping")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerPing(Resource):
    @ns.doc("update_worker_contact")
    @ns.expect(ping_parser)
    @ns.marshal_with(worker)
    def put(self, id):
        """Update a worker last contact"""
        args = ping_parser.parse_args()
        worker_dao.update_contact(id, args.get('load',''),args.get('memory',''),args.get('stats',''))
        return worker_dao.get(id)

callback_parser = api.parser()
callback_parser.add_argument('message', type=str, help='Callback message sent (idle)', location='json')

@ns.route("/<id>/callback")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerCallback(Resource):
    @ns.doc("update_worker_contact")
    @ns.expect(callback_parser)
    def put(self, id):
        """Update a worker last contact"""
        message = callback_parser.parse_args().get('message','')
        worker = worker_dao.get(id)
        if message == 'idle': 
            if not worker.permanent:
                if db.session.query(Execution).filter(Execution.status.in_(['running','pending']),
                        Execution.worker_id==worker.worker_id).count()>0:
                    log.warning(f'Worker {worker.name} called idle callback but some tasks are still running, refusing...')
                    return {'result':'still have running tasks'}
                if db.session.query(Task).filter(and_(Task.status.in_(['pending']),
                                        Task.batch==worker.batch)).count()>0:
                    log.warning(f'Worker {worker.name} called idle but some tasks are still due...')
                    return {'result':'still some work to do, lazy one!'}
                log.warning(f'Worker {worker.name} ({worker.worker_id}) called idle callback, launching destruction')
                #worker.destroy()
                create_worker_destroy_job(worker, db.session, commit=False)
                #db.session.delete(worker)
            else:
                worker.task_properties = json_module.dumps({})
            db.session.commit()
            return {'result':'ok'}
        else:
            log.warning(f'Worker {worker.name} called idle callback but has no idle command')
            return {'result':'nothing to do'}

deploy_parser = api.parser()
deploy_parser.add_argument('number', type=int, help='How many workers should be deployed', location='json')
deploy_parser.add_argument('region', type=str, help='Which provider region for worker', location='json')
deploy_parser.add_argument('provider', type=str, help='Specify the worker provider', location='json')
deploy_parser.add_argument('flavor', type=str, help='Which provider flavor/model for worker', location='json')
deploy_parser.add_argument('batch', type=str, help='Batch name (that must be shared by tasks) for worker', location='json')
deploy_parser.add_argument('concurrency', type=int, help='How many tasks should be run in parallel', location='json')
deploy_parser.add_argument('prefetch', type=int, help='How many extra tasks should be prefetched', location='json')

@ns.route("/deploy")
class WorkerDeploy(Resource):
    @ns.doc("deploy_worker_vm_and_process")
    @ns.expect(deploy_parser)
    def put(self):
        """Create and deploy one or several workers"""
        deploy_args = deploy_parser.parse_args()
        create_worker_create_job(session=db.session, **deploy_args)
        #for _ in range(deploy_args['number']):
#
        #    db.session.add(
        #        Job(target='', 
        #            action='worker_create', 
        #            args={
        #                'concurrency': deploy_args['concurrency'], 
        #                'prefetch':deploy_args['prefetch'],
        #                'flavor':deploy_args['flavor'],
        #                'region':deploy_args['region'],
        #                'provider':deploy_args['provider'],
        #                'batch':deploy_args['batch']
        #            }
        #        )
        #    )
        #db.session.commit()


        return {'result':'ok'}

class ExecutionDAO(BaseDAO):
    ObjectType = Execution
    authorized_status = EXECUTION_STATUS

    def create(self, data):
        task = task_dao.get(data['task_id'])
        if task.status not in ['pending','paused']:
            api.abort(500, f"A new execution for task {task.task_id} is not possible")
        if 'status' in data:
            if data['status']=='running':
                task.status = 'running'
            elif data['status'] in ['pending','paused']:
                task.status = 'assigned'
            else:
                api.abort(500, f"A new execution cannot be created with a status {data['status']}")
        else:
            task.status='assigned'
        return super().create(data)

    def update(self, id, data):
        execution = self.get(id)
        modified = False
        for attr, value in data.items():
            if hasattr(execution,attr): 
                if getattr(execution,attr)!=value:
                    if attr=='status':
                        try:
                            execution_update_status(execution, db.session, value, commit=False)
                            modified = True
                        except ModelException as model_exception:
                            api.abort(500, model_exception.message)
                    else:
                        setattr(execution, attr, value)
                        modified = True
            else:
                raise Exception('Error: {} has no attribute {}'.format(
                    execution.__name__, attr))
        if modified:
            execution.modification_date = datetime.utcnow()
            db.session.commit()
        return execution

    def list(self,no_output=False,limit=None,reverse=False,trunc=None,**args):
        sorting_column=Execution.execution_id
        if reverse:
            sorting_column=sorting_column.desc()
        q=Execution.query
        if args:
            task_args = {}
            if 'task_name' in args:
                task_args['name']=args['task_name']
                del(args['task_name'])
            if 'batch' in args:
                task_args['batch']=args['batch']
                del(args['batch'])
            filter = None
            if args:
                filter=process_filtering_args(Execution, args)
            if task_args:
                q = q.join(Task, Task.task_id==Execution.task_id)
                if filter:
                    filter = and_(filter, process_filtering_args(Task, task_args))
                else:
                    filter = process_filtering_args(Task, task_args)
            q=q.filter(filter)
        if no_output:
            q=q.with_entities(Execution.execution_id,
                              Execution.worker_id,
                              Execution.command, 
                              Execution.status,
                              Execution.task_id,
                              Execution.creation_date,
                              Execution.modification_date,
                              Execution.pid,
                              Execution.output_files,
                              Execution.latest)
        elif trunc:
            if IS_SQLITE:
                trunc_output = func.substr(Execution.output, -trunc, trunc).label('output')
                trunc_error = func.substr(Execution.error, -trunc, trunc).label('error')
            else:
                trunc_output = func.right(Execution.output, trunc).label('output')
                trunc_error = func.right(Execution.error, trunc).label('error')
            q=q.with_entities(Execution.execution_id,
                              Execution.worker_id,
                              Execution.command, 
                              Execution.status,
                              Execution.task_id,
                              Execution.creation_date,
                              Execution.modification_date,
                              Execution.pid,
                              Execution.output_files,
                              Execution.latest,
                              trunc_output,
                              trunc_error)
        q = q.order_by(sorting_column)
        if limit:
            q=q.limit(limit)

        return list(q.all())
 
execution_dao = ExecutionDAO()

execution = api.model('Execution', {
    'execution_id': fields.Integer(readonly=True, description='The execution unique identifier'), 
    'worker_id': fields.Integer(required=True, description='A worker unique identifier'),
    'task_id': fields.Integer(required=True, description='A task unique identifier'),
    'status': fields.String(readonly=True,
        description=f'The execution status: {", ".join(ExecutionDAO.authorized_status)}'), 
    'return_code': fields.Integer(required=False, description='The return code of the execution (when finished)'),
    'pid': fields.String(required=False, description='The process id (pid) of the execution'),
    'creation_date': fields.DateTime(readonly=True,
        description="timestamp of execution creation (on worker)"),
    'modification_date': fields.DateTime(readonly=True,
        description="timestamp of execution last modification"),
    'output': fields.String(readonly=False, required=False, description='The standard output of the execution'),
    'error': fields.String(readonly=False, required=False, description='The standard error of the execution (if any)'),
    'output_files': fields.String(readonly=True, description='A list of output files transmitted (if any)'),
    'command': fields.String(required=False, description='The command that was really launched for this execution (it case Task.execution is modified)'),
    'latest': fields.Boolean(readonly=True, description='Latest or current execution for the related task')
})

execution_plus_batch = api.model('ExecutionPlusBatch', {
    'execution_id': fields.Integer(readonly=True, description='The execution unique identifier'), 
    'command': fields.String(required=False, description='The command that was really launched for this execution (it case Task.execution is modified)'),
    'status': fields.String(readonly=True,
        description=f'The execution status: {", ".join(ExecutionDAO.authorized_status)}'), 
    'task_id': fields.Integer(required=True, description='A task unique identifier'),
    'creation_date': fields.DateTime(readonly=True,
        description="timestamp of execution creation (on worker)"),
    'modification_date': fields.DateTime(readonly=True,
        description="timestamp of execution last modification"),
    'pid': fields.String(required=False, description='The process id (pid) of the execution'),
    'output_files': fields.String(readonly=True, description='A list of output files transmitted (if any)'),
    'latest': fields.Boolean(readonly=True, description='Latest or current execution for the related task'),
    'batch': fields.String(readonly=True, description='The batch of the underlying task'),
    'taskstatus': fields.String(readonly=True, description='The status of the underlying task')
})


@ns.route("/<id>/executions")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerExecutionObject(Resource):
    @ns.doc("get_worker_executions")
    @ns.marshal_list_with(execution_plus_batch)
    def get(self, id):
        """Fetch a worker executions given the worker identifier"""
        #worker_dao.update_contact(id)
        worker_executions = db.session.query(Execution.execution_id,
                              Execution.command, 
                              Execution.status,
                              Execution.task_id,
                              Execution.creation_date,
                              Execution.modification_date,
                              Execution.pid,
                              Execution.output_files,
                              Execution.latest,
                              Task.batch,
                              label('taskstatus',Task.status)).\
                        join(Execution.task).\
                        filter(and_(Execution.worker_id==id,
                                    Execution.status.not_in(['failed','succeeded','pending'])))
        return list(worker_executions)


@ns.route("/<id>/executions/<status>")
@ns.param("id", "The worker identifier")
@ns.param("status", "Filter only executions with this status")
@ns.response(404, "Worker not found")
class WorkerExecutionFilterObject(Resource):
    @ns.doc("get_worker_executions")
    @ns.marshal_list_with(execution_plus_batch)
    def get(self, id, status):
        """Fetch a worker executions given the worker identifier and the executions status"""
        #worker_dao.update_contact(id)
        #return execution_dao.list(worker_id=id, status=status)
        worker_executions = db.session.query(Execution.execution_id,
                              Execution.command, 
                              Execution.status,
                              Execution.task_id,
                              Execution.creation_date,
                              Execution.modification_date,
                              Execution.pid,
                              Execution.output_files,
                              Execution.latest,
                              Task.batch,
                              label('taskstatus',Task.status)).\
                        join(Execution.task).\
                        filter(and_(Execution.worker_id==id,
                                    Execution.status==status))
        return list(worker_executions)

signal = api.model('Signal', {
    'execution_id': fields.Integer(readonly=True, description='The execution unique identifier'), 
    'worker_id': fields.Integer(readonly=True, description='A worker unique identifier'),
    'signal': fields.Integer(readonly=True, description='The signal to send to the execution (UNIX signal)'),
})

signal_parser = api.parser()
signal_parser.add_argument('execution_id', type=int, help='Execution identifier', location='json')
signal_parser.add_argument('signal', type=int, help='Signal to send', location='json')

@ns.route("/<id>/signals")
@ns.param("id", "The worker identifier")
@ns.response(404, "Worker not found")
class WorkerSignal(Resource):
    @ns.doc("get_worker_signals")
    @ns.marshal_list_with(signal)
    def get(self, id):
        """Fetch a worker executions given the worker identifier and the executions status"""
        signals = list(Signal.query.filter(Signal.worker_id==id))
        for sig in signals:
            db.session.delete(sig)
        db.session.commit()
        return signals
    
    @ns.doc("create_worker_signal")
    @ns.expect(signal_parser)
    def post(self, id):
        """Create a signal for the given worker about a certain execution with a certain signal"""
        args = signal_parser.parse_args()
        db.session.add(Signal(args['execution_id'], id, args['signal']))
        db.session.commit()


ns = api.namespace('executions', description='EXECUTION operations')

execution_filter = api.model('ExecutionFilter', {
    'task_id': fields.Integer(required=False,decription='A list of ids to restrict listing'),
    'status': fields.String(required=False, description="Filter with this status"),
    'latest': fields.Boolean(required=False, decription="Filter only for latest execution"),
    'task_name': fields.String(required=False, description="Filter with this task name"),
    'batch': fields.String(required=False, description="Filter with this batch"),
    'limit': fields.Integer(required=False, description="Limit results to this number"),
    'reverse': fields.Boolean(required=False, decription="Reverse sorting order, most recent executions first"),
    'no_output': fields.Boolean(required=False, decription="Do not include output and error for a lighter query"),
    'trunc': fields.Integer(required=False, description="Limit output size to this number"),
})
@ns.route('/')
class ExecutionList(Resource):
    '''Shows a list of all executions, and lets you POST to add new workers'''
    @ns.doc('list_executions')
    @ns.expect(execution_filter)
    @ns.marshal_list_with(execution)
    def get(self):
        '''List all executions'''
        return execution_dao.list(**api.payload) if api.payload else execution_dao.list()

    @ns.doc('create_execution')
    @ns.expect(execution)
    @ns.marshal_with(execution, code=201)
    def post(self):
        '''Create a new execution'''
        return execution_dao.create(api.payload), 201
    

@ns.route("/<id>")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionObject(Resource):
    @ns.doc("get_execution")
    @ns.marshal_with(execution)
    def get(self, id):
        """Fetch a execution given its identifier"""
        return execution_dao.get(id)

    @ns.doc("update_execution")
    @ns.expect(execution)
    @ns.marshal_with(execution, code=201)
    def put(self, id):
        """Update an execution"""
        return execution_dao.update(id, api.payload)


parser = api.parser()
parser.add_argument('text', type=str, help='Supplementary text', location='json')

@ns.route("/<id>/delete")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionObject(Resource):
    @ns.doc('delete_execution')
    def put(self,id):
        '''Delete an execution for this id'''
        for e in Execution.query.filter(Execution.execution_id==id):
            db.session.delete(e)
        db.session.commit()
        return {'result':'ok'}

@ns.route("/<id>/output")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionOutput(Resource):
    @ns.doc("update_execution_output")
    @ns.expect(parser)
    def put(self, id):
        """Add some data to the execution output"""
        x = execution_dao.get(id)
        args = parser.parse_args()
        execution_dao.update(id, 
            {'output':('' if x.output is None else x.output) + args['text']})
        return {'result':'Ok'}

@ns.route("/<id>/error")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionOutput(Resource):
    @ns.doc("update_execution_error")
    @ns.expect(parser)
    def put(self, id):
        """Add some data to execution error"""
        x = execution_dao.get(id)
        args = parser.parse_args()
        execution_dao.update(id, 
            {'error':('' if x.error is None else x.error) + args['text']})
        return {'result':'Ok'}

@ns.route("/<id>/output_files")
@ns.param("id", "The execution identifier")
@ns.response(404, "Execution not found")
class ExecutionOutputFiles(Resource):
    @ns.doc("update_execution_output")
    @ns.expect(parser)
    def put(self, id):
        """Update output files (files transferred from /output, i.e. results) for execution as a space separated file URL list"""
        args = parser.parse_args()
        execution_dao.update(id, 
            {'output_files':args['text']})
        return {'result':'Ok'}


ns = api.namespace('batch', description='BATCH level operations')

batch_parser = api.parser()
batch_parser.add_argument('signal', type=int, help='Optionnaly send signal to all running tasks', 
    required=False, location='json')


custom_batch = api.model('Batch', {
    'batch': fields.String(readonly=True, description='The name of the batch'), 
    'pending': fields.Integer(readonly=True,
        description='The number of task in pending state'), 
    'accepted': fields.Integer(readonly=True,
        description='The number of task in accepted state'), 
    'running': fields.Integer(readonly=True,
        description='The number of task in running state'), 
    'failed': fields.Integer(readonly=True,
        description='The number of task in succeeded state'), 
    'succeeded': fields.Integer(readonly=True,
        description='The number of task in failed state'), 
    'workers': fields.String(readonly=True, description='The workers used in that batch'),
})

@ns.route("/")
class BatchList(Resource):
    @ns.doc("list_batch")
    @ns.marshal_list_with(custom_batch)
    def get(self):
        """List the batches, their task statuses and workers"""
        if IS_SQLITE:
            worker_query='''SELECT batch,GROUP_CONCAT(name,',') FROM worker GROUP BY batch'''
        else:
            worker_query='''SELECT batch,STRING_AGG(name,',') FROM worker GROUP BY batch'''
        batch_query='''SELECT batch,status,count(task_id) FROM task GROUP BY batch,status ORDER BY batch,status'''
        batches = []
        batches_attributes = {}
        for batch,status,count in db.session.execute(batch_query):
            if batch not in batches:
                batches.append(batch)
                batches_attributes[batch]={'batch':batch}    
            batches_attributes[batch][status]=count
        for batch,workers in db.session.execute(worker_query):
            if batch in batches_attributes:
                batches_attributes[batch]['workers']=workers
        return [batches_attributes[batch] for batch in batches]
@ns.route("/<name>/stop")
@ns.param("name", "The batch name")
@ns.response(404, "Batch not found")
class BatchStop(Resource):
    @ns.doc("stop_a_batch")
    @ns.expect(batch_parser)
    def put(self, name):
        """Pause all workers for this batch - and kill current job if force is set"""
        for w in Worker.query.filter(Worker.batch==name):
            w.status = 'paused'
        args = batch_parser.parse_args()
        if args.get('signal',False):
            log.warning(f'Sending signal {args["signal"]} to executions for batch {name}')
            for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                            Execution.status=='running',
                                            Task.batch==name)):
                log.warning(f'Sending signal {args["signal"]} to execution {e.execution_id}')
                db.session.add(Signal(e.execution_id, e.worker_id, args['signal']))
        db.session.commit()
        return {'result':'Ok'}

@ns.route("/<name>/go")
@ns.param("name", "The batch name")
@ns.response(404, "Batch not found")
class BatchGo(Resource):
    @ns.doc("_re_launch_a_batch")
    @ns.expect(batch_parser)
    def put(self, name):
        """(re)set all workers affected to this batch to running"""
        for w in Worker.query.filter(Worker.batch==name):
            w.status = 'running'
        args = batch_parser.parse_args()
        if args.get('signal',False):
            log.warning(f'Sending signal {args["signal"]} to executions for batch {name}')
            for e in db.session.scalars(select(Execution).join(Execution.task).where(
                                            Execution.status=='running',
                                            Task.batch==name)):
                log.warning(f'Sending signal {args["signal"]} to execution {e.execution_id}')
                db.session.add(Signal(e.execution_id, e.worker_id, args['signal']))
        db.session.commit()
        return {'result':'Ok'}



@ns.route("/<name>")
@ns.param("name", "The batch name")
@ns.response(404, "Batch not found")
class BatchDelete(Resource):
    @ns.doc("delete_a_batch")
    def delete(self, name):
        """Delete all tasks and executions for this batch"""
        log.warning(f'Deleting batch {name}')
        delete_batch(name, session=db.session)
        return {'result':'Ok'}


ns = api.namespace('requirement', description='Task requirements management')


requirement = api.model('Requirement', {
    'requirement_id': fields.Integer(readonly=True, description='The requirement unique identifier'),
    'task_id': fields.Integer(required=True, description='The requiring task id'),
    'other_task_id': fields.Integer(required=True, description='The requirred task id'),
})

@ns.route('/')
class RequirementList(Resource):
    '''Shows a list of all requirements, and lets you POST to add new requirements'''
    @ns.doc('list_requirement')
    @ns.marshal_list_with(requirement)
    def get(self):
        '''List all requirements'''
        return requirement_dao.list()

    @ns.doc('create_requirement')
    @ns.expect(requirement)
    @ns.marshal_with(requirement, code=201)
    def post(self):
        '''Create a new requirement'''
        return requirement_dao.create(api.payload), 201

@ns.route("/<id>")
@ns.param("id", "The requirement identifier")
@ns.response(404, "Requirement not found")
class RequirementObject(Resource):
    @ns.doc("delete_requirement")
    @ns.marshal_with(requirement)
    def delete(self, id):
        """Delete a worker"""
        return requirement_dao.delete(id)



ns = api.namespace('recruiter', description='Recruiter management')

class RecruiterDAO(BaseDAO):
    ObjectType = Recruiter

    def get(self, batch, rank):
        object = Recruiter.query.get({'batch':batch, 'rank':rank})
        if object is None:
            api.abort(404, "{} batch {} rank {} doesn't exist".format(
                    self.ObjectType.__name__,batch,rank))
        return object
        
    def create(self, data):
        object = self.ObjectType(**data)
        db.session.merge(object)
        db.session.commit()
        return object

    def update(self, batch, rank, data):
        object = self.get(batch, rank)
        modified = False
        for attr, value in data.items():
            if hasattr(object,attr): 
                if getattr(object,attr)!=value:
                    setattr(object, attr, value)
                    modified = True
            else:
                api.abort(500,f'Error: {object.__name__} has no attribute {attr}')
        if modified:
            db.session.commit()
        return object

    def delete(self, batch, rank):
        object = self.get(batch,rank)
        db.session.delete(object)
        db.session.commit()
        return object
    
    def list(self, **args):
        return super().list(sorting_column='rank', **args)
    


recruiter_dao = RecruiterDAO()

recruiter = api.model('Recruiter', {
    'batch': fields.String(readonly=True, description='The target batch for recruitment'),
    'rank': fields.Integer(readonly=True, description='The rank of the recruiter for that batch (unique per batch)'),
    'tasks_per_worker': fields.Integer(required=True, description='Each time this number of pending tasks is there, trigger a recruitment'),
    'worker_flavor': fields.String(required=True, description='What flavor of worker to recruit'),
    'worker_provider': fields.String(required=False, description='From what provider the worker should be recruited'),
    'worker_region': fields.String(required=False, description='From what provider region the worker should be recruited'),
    'worker_concurrency': fields.Integer(required=True, description='Set worker concurrency to this when recruited'),
    'worker_prefetch': fields.Integer(required=False, description='Set worker prefetch to this when recruited (0 otherwise)'),
    'minimum_tasks': fields.Integer(required=False, description='Do not trigger until there is this minimum number of task (should be above tasks_per_worker)'),
    'maximum_workers': fields.Integer(required=False, description='Stop to trigger when there is this number of worker for the batch'),
})

@ns.route('/')
class RecruiterList(Resource):
    '''Shows a list of all recruiters, and lets you POST to add new recruiters'''
    @ns.doc('list_recruiters')
    @ns.marshal_list_with(recruiter)
    def get(self):
        '''List all recruiters'''
        return recruiter_dao.list(**api.payload)

    @ns.doc('create_recruiter')
    @ns.expect(recruiter)
    @ns.marshal_with(recruiter)
    def post(self):
        '''Create or replace a recruiter'''
        return recruiter_dao.create(api.payload), 201

@ns.route("/<batch>/<rank>")
@ns.param("batch", "Target batch for the recruiter")
@ns.param("rank", "Rank of the recruiter")
@ns.response(404, "Recruiter not found")
class RecruiterObject(Resource):
    @ns.doc("get_recruiter")
    @ns.marshal_with(recruiter)
    def get(self, batch, rank):
        """Fetch a execution given its identifier"""
        return recruiter_dao.get(batch, rank)

    @ns.doc("update_recruiter")
    @ns.expect(recruiter)
    @ns.marshal_with(recruiter, code=201)
    def put(self, batch, rank):
        """Update an execution"""
        return recruiter_dao.update(batch, rank, api.payload)

    @ns.doc("delete_recruiter")
    @ns.marshal_with(recruiter)
    def delete(self, batch, rank):
        """Delete a worker"""
        object=recruiter_dao.delete(batch, rank)
        return object 

ns = api.namespace('flavor', description='Flavor querying')

flavor_parser = api.parser()
flavor_parser.add_argument('min_cpu', type=int, help='Minimum number of CPUs', 
    required=False, location='json', default=0)
flavor_parser.add_argument('min_ram', type=int, help='Minimum size of RAM (memory) in Gb', 
    required=False, location='json', default=0)
flavor_parser.add_argument('min_disk', type=int, help='Minimum size of disk (harddrive) in Gb', 
    required=False, location='json', default=0)
flavor_parser.add_argument('max_eviction', type=int, 
    help=f'Maximal risk of eviction (default:{FLAVOR_DEFAULT_EVICTION})', 
    required=False, location='json', default=FLAVOR_DEFAULT_EVICTION)
flavor_parser.add_argument('limit', type=int, 
    help=f'Maximal number or answer (default:{FLAVOR_DEFAULT_LIMIT})', 
    required=False, location='json', default=FLAVOR_DEFAULT_LIMIT)
flavor_parser.add_argument('provider', type=str, 
    help=f'Specify provider', 
    required=False, location='json', default=None)
flavor_parser.add_argument('region', type=str, 
    help=f'Specify region', 
    required=False, location='json', default=None)
flavor_parser.add_argument('protofilters', type=str, 
    help=f'Add some : separated protofilters like cpu>10 or ram>=128', 
    required=False, location='json', default=None)

custom_flavor = api.model('Flavor', {
    'name': fields.String(readonly=True, description='The flavor reference'),
    'provider':  fields.String(readonly=True, description='Provider name'),
    'region':  fields.String(readonly=True, description='Provider region name'),
    'cpu': fields.Integer(readonly=True,
        description='The number of CPUs (vcores) for this flavor'), 
    'ram': fields.Float(readonly=True,
        description='The size of the RAM (memory) in Gb'), 
    'disk': fields.Float(readonly=True,
        description='The size of the disk in Gb'),
    'tags': fields.String(readonly=True, 
        description='A custom string listing some specificity of the flavor (G=GPU, M=Metal, N=NVMe)'),
    'gpu': fields.String(readonly=True,
        description='A description of the GPU (if available)'),
    'gpumem': fields.Float(readonly=True,
        description='The size of the GPU memory in Gb (if available)'),
    'cost': fields.Float(readonly=True, description='An indicative cost per hour in the unit of reference of the provider'),
    'eviction': fields.Integer(readonly=True, description='An estimate of the hourly probability of eviction (VM shut down by provider)'), 
    'available': fields.Integer(readonly=True, 
        description='An estimate of the number of VM of this size one can create in this region and for this provider')
})

@ns.route("/")
class FlavorList(Resource):
    @ns.doc("list_flavor")
    @ns.marshal_list_with(custom_flavor)
    @ns.expect(flavor_parser)
    def get(self):
        """List the batches, their task statuses and workers"""
        args = flavor_parser.parse_args()
        return find_flavor(session=db.session, 
                    min_cpu=args.min_cpu,
                    min_ram=args.min_ram, 
                    min_disk=args.min_disk,
                    max_eviction=args.max_eviction,
                    limit = args.limit,
                    provider = args.provider,
                    region = args.region,
                    protofilters = args.protofilters)



"""Changing Worker field idle_callback to permanent

Revision ID: 740850a52979
Revises: c225e89357b9
Create Date: 2024-03-17 22:43:20.712985

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '740850a52979'
down_revision = 'c225e89357b9'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('worker', schema=None) as batch_op:
        batch_op.add_column(sa.Column('permanent', sa.Boolean(), server_default="1"))
        batch_op.drop_column('idle_callback')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('worker', schema=None) as batch_op:
        batch_op.add_column(sa.Column('idle_callback', sa.VARCHAR(), nullable=True))
        batch_op.drop_column('permanent')

    # ### end Alembic commands ###



from azure.identity import ClientSecretCredential
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.subscription import SubscriptionClient
from azure.mgmt.resourcegraph import ResourceGraphClient
from azure.mgmt.resourcegraph.models import QueryRequest, QueryRequestOptions

import os
import requests
import json
import csv
from io import StringIO
import logging as log
from ...server import get_session, config
from ..generic import GenericProvider, Flavor, FlavorMetrics
from ...util import package_path
import re

GPU_RESOURCE='providers/azure/gpu.tsv'
FLAVOR_NAME_RE=re.compile(r'Standard_(?P<main>[A-Z0-9-]+)(?P<option>[a-z]+)?(_.*?)?$')
DEFAULT_OS_DISK=30

class Azure(GenericProvider):
    def __init__(self, subscription_id, client_id, client_secret, tenant_id, session,
                 regions=None, quotas=None, live=True):
        self.credentials = ClientSecretCredential(  
            client_id=client_id,  
            client_secret=client_secret,  
            tenant_id=tenant_id  
            )
        self.subscription_client = SubscriptionClient(self.credentials)
        self.compute_client = ComputeManagementClient(self.credentials, subscription_id)
        self.resourcegraph_client = ResourceGraphClient(credential=self.credentials, subscription_id=subscription_id)
        self.live = live


        if regions is None:
            self.regions = list([region.name for region in 
                                 self.subscription_client.subscriptions.list_locations(subscription_id)])
        else:
            self.regions = regions
        self.regions.sort()
        self.quotas = dict(zip(regions,quotas))
        self.session = session
        self.provider = 'azure'


            

#regions = list([region.name for region in subscription_client.subscriptions.list_locations(subscription_id)])
#regions.sort()

    def get_flavors(self):
        """Update flavors"""
        print('\nAdding GPU info')
        flavor_gpu = {}
        with open(package_path(GPU_RESOURCE),'rt',encoding='utf-8') as f:
            for entry in csv.DictReader(f,delimiter='\t'):
                flavor_gpu[entry['flavor']]=entry
        self.push('Updating Azure flavors')            
        seen_flavors = []
        flavors = []
        flavor_metrics = {}
        for region in self.regions:
            self.push('.')
            try:
                for flavor in self.compute_client.virtual_machine_sizes.list(location=region):
                    # the flavor name must contain a small s to be compatible with premium storage
                    m=FLAVOR_NAME_RE.match(flavor.name)
                    if not m:
                        # we know that basic are not accepted no need to shout for such
                        if 'Basic' not in flavor.name:
                            self.push(f'<Not a standard flavor {flavor.name}>')
                        continue
                    options=m.groupdict()['option']
                    if options is None or 's' not in options:
                        # no premium storage
                        continue
                    if 'p' in options:
                        # arm64 not supported yet
                        continue
                    if flavor.name not in seen_flavors:
                        try:
                            f=Flavor(
                                name=flavor.name,
                                provider=self.provider,
                                cpu=flavor.number_of_cores,
                                ram=flavor.memory_in_mb/1024,
                                disk=max(DEFAULT_OS_DISK,flavor.resource_disk_size_in_mb/1024), # do not use flavor.os_disk_size_in_mb, this is a maximal optional size not the real size
                                bandwidth=1,
                                tags='G' if flavor.name.startswith('Standard_N') else '')
                            if flavor.name in flavor_gpu:
                                f.gpu=flavor_gpu[flavor.name]['gpu']
                                f.gpumem=int(flavor_gpu[flavor.name]['gpumem'])
                        except:
                            log.exception('Flavor ill created')
                            self.push('!')
                        flavors.append(f)
                        seen_flavors.append(flavor.name)
                    try:
                        flavor_metrics[flavor.name,region]=FlavorMetrics(
                            flavor_name=flavor.name,
                            provider=self.provider,
                            region_name=region
                        )
                    except:
                        log.exception('FlavorMetrics ill created')
                        self.push('!')
            except:
                pass
        self.update_flavors(flavors=flavors)
        self.__flavor_metrics__ = flavor_metrics

    def get_metrics(self):
        # add price
        self.push('\nFinding prices')
        query="https://prices.azure.com/api/retail/prices?currencyCode='EUR'&$filter= serviceName eq 'Virtual Machines' and priceType eq 'Consumption' and contains(meterName, 'Spot')"
        query += ' and (' + ' or '.join([f"armRegionName eq '{region}'" for region in self.regions]) + ')'
        retain_metrics = []
        while query:
            self.push('.')
            data = requests.get(query).json()
            for item in data['Items']:
                flavor = item['armSkuName']
                region = item['armRegionName']
                price = item['unitPrice']

                if (flavor,region) in self.__flavor_metrics__:
                    retain_metrics.append( (flavor,region) )
                    metrics = self.__flavor_metrics__[flavor, region]
                    try:
                        metrics.cost = price
                    except:
                        self.push('!')


            query=data.get('NextPageLink',None)
    
        self.__flavor_metrics__ = {k:v for k,v in self.__flavor_metrics__.items() if k in retain_metrics}
        flavors_lower = { tuple((k.lower() for k in key)):value for key,value in self.__flavor_metrics__.items() }

        # add eviction
        self.push('\nFinding eviction rates')
        

        skip_token = True
        full_metric_pkeys = []
        while skip_token:
            self.push('.')
            if skip_token is True:
                options=None
            else:
                options=QueryRequestOptions(skip_token=skip_token)
            query = QueryRequest(
                    query="spotresources \
        | where type == 'microsoft.compute/skuspotevictionrate/location' \
        | project skuName=tostring(sku.name), location, evictionRate=properties.evictionRate, price=properties.unitPrice \
        | order by skuName", options=options
                )
            query_response = self.resourcegraph_client.resources(query, )


            for item in query_response.data:
                if (item['skuName'],item['location']) in flavors_lower:
                    evictionRate=item['evictionRate'].split('-')[-1].strip('+')
                    eviction=int(evictionRate)
                    try:
                        flavors_lower[(item['skuName'],item['location'])].eviction=eviction
                    except:
                        self.push('!')
                    if (item['skuName'],item['location']) not in full_metric_pkeys:
                        full_metric_pkeys.append((item['skuName'],item['location']))


            skip_token=query_response.skip_token

        full_metrics = [flavors_lower[fmpk] for fmpk in full_metric_pkeys]
        self.update_flavor_metrics(metrics=full_metrics)
        self.push('\n')


def run():
    session = get_session()
    regions = config.AZURE_REGIONS.split()
    quotas = config.AZURE_CPUQUOTAS.split()
    if not regions or not quotas:
        raise RuntimeError('AZURE_REGIONS and AZURE_CPUQUOTAS *must* be set in /etc/scitq.conf for the updater to work')
    subscription_id = config.AZURE_SUBSCRIPTION_ID
    client_id = config.AZURE_CLIENT_ID
    client_secret = config.AZURE_SECRET 
    tenant_id = config.AZURE_TENANT

    updater=Azure(subscription_id=subscription_id, client_id=client_id, client_secret=client_secret, tenant_id=tenant_id,
                  session=session, regions=regions, quotas=quotas)
    updater.get_flavors()
    updater.get_metrics()
